from itertools import cycle
from unittest.mock import patch

import numpy as np
import pytest

from zhinst.toolkit import Waveforms
from zhinst.core import compile_seqc


@pytest.fixture()
def generator(data_dir, mock_connection, shfqa):

    json_path = data_dir / "nodedoc_awg_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.awgModule.return_value.listNodesJSON.return_value = (
        nodes_json
    )

    yield shfqa.qachannels[0].generator


def test_enable_sequencer(shfqa, mock_connection):
    shfqa.qachannels[0].generator.enable_sequencer(single=True)
    mock_connection.return_value.set.assert_called_with(
        "/dev1234/qachannels/0/generator/single", 1
    )
    mock_connection.return_value.syncSetInt.assert_called_with(
        "/dev1234/qachannels/0/generator/enable", 1
    )

    yield shfqa.qachannels[0].generator


def test_load_sequencer_program_empty_string_none(generator):
    with pytest.raises(RuntimeError):
        generator.load_sequencer_program("")
    with pytest.raises(RuntimeError):
        generator.load_sequencer_program(None)


def test_wait_done(mock_connection, shfqa, generator):
    single = 0
    enable = iter([])

    def get_int_side_effect(node):
        if node.upper() == "/DEV1234/QACHANNELS/0/GENERATOR/SINGLE":
            return single
        if node.upper() == "/DEV1234/QACHANNELS/0/GENERATOR/ENABLE":
            return next(enable)

    mock_connection.return_value.getInt.side_effect = get_int_side_effect

    # if not single mode this function throws a RuntimeError
    with pytest.raises(RuntimeError) as e_info:
        shfqa.qachannels[0].generator.wait_done()

    # already finished
    single = 1
    enable = iter([0] * 2)
    shfqa.qachannels[0].generator.wait_done()
    # finishes in time
    single = 1
    enable = iter([1] * 3 + [0] * 2)
    shfqa.qachannels[0].generator.wait_done()
    # don't finish
    single = 1
    enable = cycle([1])
    with pytest.raises(TimeoutError) as e_info:
        shfqa.qachannels[0].generator.wait_done(timeout=0.1)


def test_load_sequencer_program(shfqa, generator, mock_connection):
    # everything ok
    elf, info_original = compile_seqc("setTrigger(1);", "SHFQA4", [], 0)
    info = shfqa.qachannels[0].generator.load_sequencer_program("setTrigger(1);")
    mock_connection.return_value.set.assert_called_once()
    assert (
        mock_connection.return_value.set.call_args[0][0]
        == "/dev1234/qachannels/0/generator/elf/data"
    )
    assert mock_connection.return_value.set.call_args[0][1] == elf
    assert info == info_original


def test_write_to_waveform_memory(shfqa, generator, mock_connection):

    waveforms = Waveforms()
    waveforms[0] = np.ones(1000)
    waveforms[1] = np.ones(1000, dtype=np.complex128)
    waveforms[2] = (np.ones(1000), np.ones(1000))

    waveforms_long = Waveforms()
    waveforms_long[1000] = np.zeros(1000)

    generator.write_to_waveform_memory(waveforms)
    complex_wave = np.empty(1000, dtype=np.complex128)
    complex_wave.real = np.ones(1000)
    complex_wave.imag = np.ones(1000)
    complex_wave = complex_wave

    assert mock_connection.return_value.set.call_args[0][0][0] == (
        "/dev1234/qachannels/0/generator/clearwave",
        1,
    )
    assert (
        mock_connection.return_value.set.call_args[0][0][1][0]
        == "/dev1234/qachannels/0/generator/waveforms/0/wave"
    )
    assert all(
        mock_connection.return_value.set.call_args[0][0][1][1]
        == np.ones(1000, dtype=np.complex128)
    )
    assert (
        mock_connection.return_value.set.call_args[0][0][2][0]
        == "/dev1234/qachannels/0/generator/waveforms/1/wave"
    )
    assert all(
        mock_connection.return_value.set.call_args[0][0][2][1]
        == np.ones(1000, dtype=np.complex128)
    )
    assert (
        mock_connection.return_value.set.call_args[0][0][3][0]
        == "/dev1234/qachannels/0/generator/waveforms/2/wave"
    )
    assert all(mock_connection.return_value.set.call_args[0][0][3][1] == complex_wave)

    shfqa.qachannels[0].generator.write_to_waveform_memory(
        waveforms, clear_existing=False
    )
    assert (
        mock_connection.return_value.set.call_args[0][0][0][0]
        == "/dev1234/qachannels/0/generator/waveforms/0/wave"
    )
    assert all(
        mock_connection.return_value.set.call_args[0][0][0][1]
        == np.ones(1000, dtype=np.complex128)
    )
    assert (
        mock_connection.return_value.set.call_args[0][0][1][0]
        == "/dev1234/qachannels/0/generator/waveforms/1/wave"
    )
    assert all(
        mock_connection.return_value.set.call_args[0][0][1][1]
        == np.ones(1000, dtype=np.complex128)
    )
    assert (
        mock_connection.return_value.set.call_args[0][0][2][0]
        == "/dev1234/qachannels/0/generator/waveforms/2/wave"
    )
    assert all(mock_connection.return_value.set.call_args[0][0][2][1] == complex_wave)

    shfqa.qachannels[0].generator.write_to_waveform_memory(
        {0: np.ones(1000)}, clear_existing=True
    )
    assert mock_connection.return_value.set.call_args[0][0][0] == (
        "/dev1234/qachannels/0/generator/clearwave",
        1,
    )
    assert (
        mock_connection.return_value.set.call_args[0][0][1][0]
        == "/dev1234/qachannels/0/generator/waveforms/0/wave"
    )
    assert all(
        mock_connection.return_value.set.call_args[0][0][1][1]
        == np.ones(1000, dtype=np.complex128)
    )

    with pytest.raises(RuntimeError) as e_info:
        shfqa.qachannels[0].generator.write_to_waveform_memory(waveforms_long)


def test_read_from_waveform_memory(shfqa, mock_connection):

    result = shfqa.qachannels[0].generator.read_from_waveform_memory()
    assert len(result) == 0
    mock_connection.return_value.get.assert_called_once_with(
        "/dev1234/qachannels/0/generator/waveforms/*/wave",
        settingsonly=False,
        flat=True,
    )

    result = shfqa.qachannels[0].generator.read_from_waveform_memory([0, 3])
    mock_connection.return_value.get.assert_called_with(
        "/dev1234/qachannels/0/generator/waveforms/0/wave,/dev1234/qachannels/0/generator/waveforms/3/wave",
        settingsonly=False,
        flat=True,
    )

    mock_connection.return_value.get.return_value = {
        "/dev1234/qachannels/0/generator/waveforms/0/wave": [
            {"vector": np.ones(1000, dtype=np.complex128)}
        ],
        "/dev1234/qachannels/0/generator/waveforms/1/wave": [
            {"vector": np.ones(1000, dtype=np.complex128)}
        ],
    }
    result = shfqa.qachannels[0].generator.read_from_waveform_memory()
    assert len(result) == 2
    np.allclose(result[0][0], np.ones(1000, dtype=np.complex128))
    assert not result[0][1]
    assert not result[0][2]
    np.allclose(result[1][0], np.ones(1000, dtype=np.complex128))


def test_configure_sequencer_triggering(generator, mock_connection):
    with patch("zhinst.toolkit.driver.devices.shfqa.utils", autospec=True) as utils:
        generator.configure_sequencer_triggering(
            aux_trigger="fobarob", play_pulse_delay=0.0001
        )
        utils.configure_sequencer_triggering.assert_called_with(
            mock_connection.return_value,
            "DEV1234",
            0,
            aux_trigger="fobarob",
            play_pulse_delay=0.0001,
        )


def test_available_aux_trigger_inputs(shfqa):
    assert shfqa.qachannels[0].generator.available_aux_trigger_inputs == [
        "chan0trigin0",
        "chan0trigin1",
        "chan1trigin0",
        "chan1trigin1",
        "chan2trigin0",
        "chan2trigin1",
        "chan3trigin0",
        "chan3trigin1",
        "chan0seqtrig0",
        "chan1seqtrig0",
        "chan2seqtrig0",
        "chan3seqtrig0",
        "chan0rod",
        "chan1rod",
        "chan2rod",
        "chan3rod",
        "swtrig0",
    ]



from unittest.mock import patch

import pytest

from zhinst.toolkit import SHFQAChannelMode
from zhinst.toolkit.driver.devices.shfqa import Generator, QAChannel, Readout, SHFScope
from zhinst.toolkit.driver.devices.shfsg import AWG, Node, SGChannel


def test_repr(shfqc):
    assert repr(shfqc) == "SHFQC(SHFQC,DEV1234)"


def test_factory_reset(shfqc):
    # factory reset not yet implemented
    with pytest.warns(RuntimeWarning) as record:
        shfqc.factory_reset()


def test_start_continuous_sw_trigger(mock_connection, shfqc):
    with patch("zhinst.toolkit.driver.devices.shfqa.utils", autospec=True) as utils:
        shfqc.start_continuous_sw_trigger(num_triggers=10, wait_time=20.0)
        utils.start_continuous_sw_trigger.assert_called_once_with(
            mock_connection.return_value, "DEV1234", num_triggers=10, wait_time=20.0
        )


def test_max_qubits_per_channel(shfqc):
    with patch("zhinst.toolkit.driver.devices.shfqa.utils", autospec=True) as utils:
        shfqc.max_qubits_per_channel
        # cached property -> second call should not call device_utils
        shfqc.max_qubits_per_channel
        utils.max_qubits_per_channel.assert_called_once()


def test_qachannels(shfqc):
    assert len(shfqc.qachannels) == 1
    assert isinstance(shfqc.qachannels[0], QAChannel)


def test_scopes(shfqc):
    assert len(shfqc.scopes) == 1
    assert isinstance(shfqc.scopes[0], SHFScope)


def test_qa_configure_channel(mock_connection, shfqc):
    with patch("zhinst.toolkit.driver.devices.shfqa.utils", autospec=True) as utils:
        shfqc.qachannels[0].configure_channel(
            input_range=10,
            output_range=20,
            center_frequency=30.0,
            mode=SHFQAChannelMode.READOUT,
        )
        utils.configure_channel.assert_called_once_with(
            mock_connection.return_value,
            "DEV1234",
            0,
            input_range=10,
            output_range=20,
            center_frequency=30.0,
            mode="readout",
        )


def test_qa_generator(shfqc, data_dir, mock_connection):
    json_path = data_dir / "nodedoc_awg_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.awgModule.return_value.listNodesJSON.return_value = (
        nodes_json
    )
    assert isinstance(shfqc.qachannels[0].generator, Generator)
    assert shfqc.qachannels[0].generator.root == shfqc.root
    assert shfqc.qachannels[0].generator.raw_tree == shfqc.raw_tree + (
        "qachannels",
        "0",
        "generator",
    )


def test_qa_readout(shfqc):
    assert isinstance(shfqc.qachannels[0].readout, Readout)
    assert shfqc.qachannels[0].readout.root == shfqc.root
    assert shfqc.qachannels[0].readout.raw_tree == shfqc.raw_tree + (
        "qachannels",
        "0",
        "readout",
    )


def test_sgchannels(shfqc, mock_connection):
    assert len(shfqc.sgchannels) == 6
    assert isinstance(shfqc.sgchannels[0], SGChannel)
    # Wildcards nodes will be converted into normal Nodes
    assert not isinstance(shfqc.sgchannels["*"], SGChannel)
    assert isinstance(shfqc.sgchannels["*"], Node)


def test_sg_awg(shfqc, data_dir, mock_connection):
    json_path = data_dir / "nodedoc_awg_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.awgModule.return_value.listNodesJSON.return_value = (
        nodes_json
    )
    assert isinstance(shfqc.sgchannels[0].awg, AWG)
    assert shfqc.sgchannels[0].awg.root == shfqc.root
    assert shfqc.sgchannels[0].awg.raw_tree == shfqc.raw_tree + (
        "sgchannels",
        "0",
        "awg",
    )


def test_sg_awg_modulation_freq(mock_connection, shfqc, data_dir):
    json_path = data_dir / "nodedoc_awg_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.awgModule.return_value.listNodesJSON.return_value = (
        nodes_json
    )

    mock_connection.return_value.getInt.return_value = 0
    shfqc.sgchannels[0].awg_modulation_freq()
    mock_connection.return_value.getInt.assert_called_with(
        "/dev1234/sgchannels/0/sines/0/oscselect"
    )
    mock_connection.return_value.getDouble.assert_called_with(
        "/dev1234/sgchannels/0/oscs/0/freq"
    )
    mock_connection.return_value.getInt.return_value = 1
    shfqc.sgchannels[0].awg_modulation_freq()
    mock_connection.return_value.getInt.assert_called_with(
        "/dev1234/sgchannels/0/sines/0/oscselect"
    )
    mock_connection.return_value.getDouble.assert_called_with(
        "/dev1234/sgchannels/0/oscs/1/freq"
    )


def test_awg_configure_marker_and_trigger(data_dir, mock_connection, shfqc):
    json_path = data_dir / "nodedoc_awg_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.awgModule.return_value.listNodesJSON.return_value = (
        nodes_json
    )

    with patch("zhinst.toolkit.driver.devices.shfsg.utils", autospec=True) as utils:
        shfqc.sgchannels[0].awg.configure_marker_and_trigger(
            trigger_in_source="test1",
            trigger_in_slope="test2",
            marker_out_source="test3",
        )
        utils.configure_marker_and_trigger.assert_called_once_with(
            mock_connection.return_value,
            "DEV1234",
            0,
            trigger_in_source="test1",
            trigger_in_slope="test2",
            marker_out_source="test3",
        )

    assert shfqc.sgchannels[0].awg.available_trigger_inputs == [
        "trigin0",
        "trigin1",
        "trigin2",
        "trigin3",
        "trigin4",
        "trigin5",
        "trigin6",
        "trigin7",
        "inttrig",
    ]

    assert shfqc.sgchannels[0].awg.available_trigger_slopes == [
        "level_sensitive",
        "rising_edge",
        "falling_edge",
        "both_edges",
    ]

    assert shfqc.sgchannels[0].awg.available_marker_outputs == [
        "awg_trigger0",
        "awg_trigger1",
        "awg_trigger2",
        "awg_trigger3",
        "output0_marker0",
        "output0_marker1",
        "output1_marker0",
        "output1_marker1",
        "trigin0",
        "trigin1",
        "trigin2",
        "trigin3",
        "trigin4",
        "trigin5",
        "trigin6",
        "trigin7",
        "low",
        "high",
    ]


def test_configure_channel(mock_connection, shfqc):
    with patch("zhinst.toolkit.driver.devices.shfsg.utils", autospec=True) as utils:
        shfqc.sgchannels[0].configure_channel(
            enable=True, output_range=111, center_frequency=20.4, rf_path=False
        )
        utils.configure_channel.assert_called_once_with(
            mock_connection.return_value,
            "DEV1234",
            0,
            enable=1,
            output_range=111,
            center_frequency=20.4,
            rflf_path=0,
        )


def test_configure_pulse_modulation(mock_connection, shfqc):
    with patch("zhinst.toolkit.driver.devices.shfsg.utils", autospec=True) as utils:
        shfqc.sgchannels[0].configure_pulse_modulation(
            enable=True,
            osc_index=5,
            osc_frequency=22.0,
            phase=856.6789,
            global_amp=7.0,
            gains=(3.0, -1.0, 5.0, 1.0),
            sine_generator_index=8,
        )
        utils.configure_pulse_modulation.assert_called_once_with(
            mock_connection.return_value,
            "DEV1234",
            0,
            enable=1,
            osc_index=5,
            osc_frequency=22.0,
            phase=856.6789,
            global_amp=7.0,
            gains=(3.0, -1.0, 5.0, 1.0),
            sine_generator_index=8,
        )


def test_configure_sine_generation(mock_connection, shfqc):
    with patch("zhinst.toolkit.driver.devices.shfsg.utils", autospec=True) as utils:
        shfqc.sgchannels[0].configure_sine_generation(
            enable=True,
            osc_index=5,
            osc_frequency=22.0,
            phase=856.6789,
            gains=(3.0, -1.0, 5.0, 1.0),
            sine_generator_index=8,
        )
        utils.configure_sine_generation.assert_called_once_with(
            mock_connection.return_value,
            "DEV1234",
            0,
            enable=1,
            osc_index=5,
            osc_frequency=22.0,
            phase=856.6789,
            gains=(3.0, -1.0, 5.0, 1.0),
            sine_generator_index=8,
        )



from itertools import cycle
from unittest.mock import patch

import numpy as np
import pytest

from zhinst.toolkit import Sequence, Waveforms


def test_assignment():
    seq1 = Sequence("Test", constants={"TEST": 1}, waveforms=Waveforms())

    assert seq1.code == "Test"
    seq1.code = "new value"
    assert seq1.code == "new value"

    assert seq1.constants["TEST"] == 1
    seq1.constants["TEST"] = 2
    assert seq1.constants["TEST"] == 2
    seq1.constants["VALUE"] = 100
    assert seq1.constants["VALUE"] == 100

    assert list(seq1.waveforms.keys()) == []
    waves = Waveforms()
    waves[0] = np.ones(100)
    waves[3] = np.ones(100)
    seq1.waveforms = waves
    assert list(seq1.waveforms.keys()) == [0, 3]


def test_to_string():
    waveforms = Waveforms()
    waveforms[0] = (0.5 * np.ones(1008), -0.2 * np.ones(1008), np.ones(1008))

    sequencer = Sequence()
    sequencer.constants["PULSE_WIDTH"] = 10e-9  # ns
    sequencer.waveforms = waveforms
    sequencer.code = """\
// Hello World
repeat(5)
..."""

    assert sequencer.to_string() == str(sequencer)
    assert (
        sequencer.to_string()
        == """\
// Constants
const PULSE_WIDTH = 1e-08;
// Waveforms declaration
assignWaveIndex(placeholder(1008, true, false), placeholder(1008, false, false), 0);
// Hello World
repeat(5)
..."""
    )

    sequencer.constants = {}
    assert (
        sequencer.to_string()
        == """\
// Waveforms declaration
assignWaveIndex(placeholder(1008, true, false), placeholder(1008, false, false), 0);
// Hello World
repeat(5)
..."""
    )

    assert (
        sequencer.to_string(waveform_snippet=False)
        == """\
// Hello World
repeat(5)
..."""
    )
    sequencer.waveforms = None
    assert (
        sequencer.to_string()
        == """\
// Hello World
repeat(5)
..."""
    )


def test_to_string_existing_constant():
    sequencer = Sequence()
    sequencer.constants["PULSE_WIDTH"] = 10e-9
    sequencer.constants["TEST"] = 10
    sequencer.code = """\
// Hello World
repeat(5)
{
    const PULSE_WIDTH = 5;

    PULSE_WIDTH
}
..."""

    assert (
        sequencer.to_string()
        == """\
// Constants
const TEST = 10;
// Hello World
repeat(5)
{
    const PULSE_WIDTH = 1e-08;

    PULSE_WIDTH
}
..."""
    )



import json
from array import array
from collections import OrderedDict
from unittest import mock
from unittest.mock import patch

import pytest

from zhinst.toolkit.driver.nodes.command_table_node import CommandTableNode
from zhinst.toolkit.driver.devices.shfsg import SHFSG


@pytest.fixture()
def command_table_node(shfsg_no_ct_schema):
    yield CommandTableNode(
        shfsg_no_ct_schema.root,
        ("sgchannels", "0", "awg", "commandtable"),
        device_type="shfsg",
    )


@pytest.fixture()
def shfsg_no_ct_schema(data_dir, mock_connection, session):
    json_path = data_dir / "nodedoc_dev1234_shfsg.json"
    with json_path.open("r", encoding="UTF-8") as file:

        nodes_json = file.read()
        f = json.loads(nodes_json)
        f.pop("/dev1234/sgchannels/0/awg/commandtable/schema")
        nodes_json = json.dumps(f)
    mock_connection.return_value.listNodesJSON.return_value = nodes_json

    mock_connection.return_value.getString.return_value = ""
    yield SHFSG("DEV1234", "SHFSG8", session)


def test_attributes_init_node(command_table_node):
    assert command_table_node._device_type == "shfsg"
    assert command_table_node.raw_tree == ("sgchannels", "0", "awg", "commandtable")


def test_ct_schema_load_from_device(shfsg, mock_connection):
    ct_schema = '{\n  "$schema": "https://json-schema.org/draft-04/schema#",\n  "title": "AWG Command Table Schema",\n}\n'
    d = {
        "/dev1234/sgchannels/0/awg/commandtable/schema": [
            {"timestamp": 31066847852080, "flags": 0, "vector": ct_schema}
        ]
    }
    schema_rturn = OrderedDict(d)
    mock_connection.return_value.get.return_value = schema_rturn
    assert shfsg.sgchannels[0].awg.commandtable.schema() == ct_schema


def test_correct_ct_node_schema_loaded(shfsg_no_ct_schema):
    mock_json = {"test ": 123}
    with patch(
        "builtins.open", mock.mock_open(read_data=json.dumps(mock_json))
    ) as mock_open:
        assert (
            shfsg_no_ct_schema.sgchannels[0].awg.commandtable.load_validation_schema()
            == mock_json
        )


@pytest.mark.parametrize(
    "payload, validate",
    [
        (
            {
                "$schema": "https://json-schema.org/draft-04/schema#",
                "header": {"version": "1.1", "userString": "Test string"},
                "table": [],
            },
            True,
        ),
        (
            {
                "$schema": "https://json-schema.org/draft-04/schema#",
                "header": {"version": "1.1", "userString": "Test string"},
                "table": [],
            },
            False,
        ),
        (
            json.dumps(
                {
                    "$schema": "https://json-schema.org/draft-04/schema#",
                    "header": {"version": "1.1", "userString": "Test string"},
                    "table": [],
                }
            ),
            False,
        ),
    ],
)
def test_ct_node_upload_to_device(
    payload, validate, mock_connection, command_table_node
):
    mock_connection.return_value.set = mock.Mock(side_effect=RuntimeError)
    command_table_node.check_status = mock.Mock(return_value="")
    command_table_node.upload_to_device(payload, validate=validate)
    mock_connection.return_value.setVector.assert_called_with(
        "/dev1234/sgchannels/0/awg/commandtable/data",
        (
            '{"$schema": "https://json-schema.org/draft-04/schema#", '
            '"header": {"version": "1.1", "userString": "Test string"}, "table": []}'
        ),
    )


@pytest.fixture
def shfsg_ct_node(mock_connection, shfsg_no_ct_schema):
    schema = {
        "$schema": "https://json-schema.org/draft-04/schema#",
        "header": {"version": "1.1", "userString": "Test string"},
        "table": [],
    }
    mock_connection.return_value.get.return_value = OrderedDict(
        [
            (
                "/dev1234/sgchannels/0/awg/commandtable/data",
                {
                    "timestamp": array("q", [3880906635]),
                    "value": [schema],
                },
            )
        ]
    )
    return CommandTableNode(
        shfsg_no_ct_schema.root,
        ("sgchannels", "0", "awg", "commandtable"),
        device_type="shfsg",
    )


def test_ct_node_load_from_device_correct_data(shfsg_ct_node, mock_connection):
    ct = shfsg_ct_node.load_from_device()
    assert ct.as_dict() == {
        "$schema": "https://json-schema.org/draft-04/schema#",
        "header": {"version": "1.1", "userString": "Test string"},
        "table": [],
    }


def test_ct_node_load_from_device_correct_node(shfsg_ct_node, mock_connection):
    shfsg_ct_node.load_from_device()
    mock_connection.return_value.get.assert_called_with(
        "/dev1234/sgchannels/0/awg/commandtable/data", settingsonly=False, flat=True
    )


def test_ct_node_ct_not_supported(command_table_node):
    command_table_node._device_type = "shfasdddsssss132"
    with pytest.raises(KeyError) as e_info:
        command_table_node.load_validation_schema()


def test_ct_node_check_status_3bit_up(command_table_node, mock_connection):
    mock_connection.return_value.getInt.return_value = 8
    with pytest.raises(RuntimeError):
        command_table_node.check_status()


def test_ct_node_check_status_1bit_up(command_table_node, mock_connection):
    mock_connection.return_value.getInt.return_value = 1
    assert command_table_node.check_status() is True

    mock_connection.return_value.getInt.return_value = 2
    assert command_table_node.check_status() is False


def test_ct_node_status_called(command_table_node, mock_connection):
    mock_connection.return_value.getInt.return_value = 1
    command_table_node.check_status()
    mock_connection.return_value.getInt.assert_called_with(
        "/dev1234/sgchannels/0/awg/commandtable/status"
    )



"""DAQ Module."""

import logging
import typing as t
from collections import namedtuple

import numpy as np
from zhinst.core import DataAcquisitionModule as ZIDAQModule

from zhinst.toolkit.driver.modules.base_module import BaseModule
from zhinst.toolkit.nodetree import Node

if t.TYPE_CHECKING:
    from zhinst.toolkit.session import Session

logger = logging.getLogger(__name__)
DAQResult = namedtuple("DAQResult", ["header", "value", "time", "frequency", "shape"])


class DAQModule(BaseModule):
    """Data Acquisition Module.

    The Data Acquisition Module corresponds to the Data Acquisition tab of the
    LabOne User Interface. It enables the user to record and align time and
    frequency domain data from multiple instrument signal sources at a defined
    data rate. The data may be recorded either continuously or in bursts based
    upon trigger criteria analogous to the functionality provided by laboratory
    oscilloscopes.

    For a complete documentation see the LabOne user manual
    https://docs.zhinst.com/labone_programming_manual/data_acquisition_module.html

    Args:
        daq_module: Instance of the core DAQ module.
        session: Session to the Data Server.
    """

    def __init__(self, daq_module: ZIDAQModule, session: "Session"):
        super().__init__(daq_module, session)
        self.root.update_nodes(
            {
                "/triggernode": {
                    "GetParser": self._get_node,
                    "SetParser": self._set_node,
                }
            },
            raise_for_invalid_node=False,
        )

    def _get_node(self, node: str) -> t.Union[Node, str]:
        """Convert a raw node string into a toolkit node.

        Overwrites the base_module._get_node function and adds the support for
        the signal subscription options.

        Args:
            node (str): Raw node string

        Returns:
            Node: Toolkit node. (if the node can not be converted the raw node
                string is returned)
        """
        try:
            return self._session.raw_path_to_node(node.replace(".", "/"), module=self)
        except (KeyError, RuntimeError):
            logger.error(
                f"Could not resolve {node} into a node of the sweeper module or "
                " a connected device."
            )
            return node

    @staticmethod
    def _set_node(signal: t.Union[Node, str]) -> str:
        """Convert a toolkit node into a raw node string.

        Support the signal subscription options.

        Args:
            signal: A node

        Returns:
            str: A raw string representation of Node
        """
        try:
            node = signal.node_info.path  # type: ignore
        except AttributeError:
            node = signal
        sample_pos = node.find("sample")
        if sample_pos != -1:
            node = node[:sample_pos] + node[sample_pos:].replace("/", ".")
        return node

    def execute(self) -> None:
        """Start the data acquisition.

        After that command any trigger will start the measurement.

        Subscription or unsubscription is not possible until the
        acquisition is finished.
        """
        self._raw_module.execute()

    def subscribe(self, signal: Node) -> None:
        """Subscribe to a node.

        The node can either be a node of this module or of a connected device.

        Support the signal subscription options of the daq module.
        For more information on the signal options see:
        https://docs.zhinst.com/labone_programming_manual/data_acquisition_module.html

        Args:
            signal: A node that should be subscribed to.
        """
        self._raw_module.subscribe(self._set_node(signal))

    def unsubscribe(self, signal: Node) -> None:
        """Unsubscribe from a node.

        The node can either be a node of this module or of a connected device.

        Args:
            signal: A node that should be unsubscribed from.
        """
        self._raw_module.unsubscribe(self._set_node(signal))

    def read(
        self, *, raw: bool = False, clk_rate: float = 60e6
    ) -> t.Dict[t.Union[Node, str], t.Union[t.List[DAQResult], t.Any]]:
        """Read the acquired data from the module.

        The data is split into bursts.

        Args:
            raw: Flag if the acquired data from the subscribed device
                device nodes should be converted into the DAQResult format
                (raw = False) or not. (default = False)
            clk_rate: Clock rate [Hz] for converting the timestamps. Only
                applies if the raw flag is reset.

        Returns:
            Result of the burst grouped by the signals.
        """
        raw_result = self._raw_module.read(flat=True)
        results = {}
        for node, node_results_raw in raw_result.items():
            if not raw and isinstance(node_results_raw[0], dict):
                node_results = []
                for result_dict in node_results_raw:
                    if "fft" in node:
                        bin_count = len(result_dict.get("value")[0])
                        bin_resolution = result_dict.get("header")["gridcoldelta"]
                        frequency = np.arange(bin_count)
                        bandwidth = bin_resolution * len(frequency)
                        frequency = frequency * bin_resolution
                        if "xiy" in node:
                            frequency = (
                                frequency - bandwidth / 2.0 + bin_resolution / 2.0
                            )
                        node_results.append(
                            DAQResult(
                                result_dict.get("header", {}),
                                result_dict.get("value"),
                                None,
                                frequency,
                                result_dict.get("value").shape,
                            )
                        )
                    else:
                        timestamp = result_dict.get("timestamp")
                        node_results.append(
                            DAQResult(
                                result_dict.get("header", {}),
                                result_dict.get("value"),
                                (timestamp[0] - timestamp[0][0]) / clk_rate,
                                None,
                                result_dict.get("value").shape,
                            )
                        )
                node_results_raw = node_results
            results[self._get_node(node)] = node_results_raw
        return results



# Copyright (C) 2020 Zurich Instruments
#
# This software may be modified and distributed under the terms
# of the MIT license. See the LICENSE file for details.
"""The Zurich Instruments Toolkit (zhinst-toolkit).

This package is a collection of Python tools for high level device
control. Based on the native interface to Zurich Instruments LabOne,
they offer an easy and user-friendly way to control Zurich Instruments
devices. It  is tailored to control multiple instruments together,
especially for device management and multiple AWG distributed control.

The Toolkit forms the basis for instrument drivers used in QCoDeS and
Labber. It comes in the form of a package compatible with Python 3.7+.
"""
from zhinst.toolkit.command_table import CommandTable
from zhinst.toolkit.interface import AveragingMode, SHFQAChannelMode
from zhinst.toolkit.session import PollFlags, Session
from zhinst.toolkit.waveform import Waveforms
from zhinst.toolkit.sequence import Sequence

try:
    from zhinst.toolkit._version import version as __version__
except ModuleNotFoundError:
    pass

__all__ = [
    "CommandTable",
    "AveragingMode",
    "SHFQAChannelMode",
    "PollFlags",
    "Session",
    "Waveforms",
    "Sequence",
]



"""Module for all device drivers."""
import typing as t

from zhinst.toolkit.driver.devices.base import BaseInstrument
from zhinst.toolkit.driver.devices.hdawg import HDAWG
from zhinst.toolkit.driver.devices.pqsc import PQSC
from zhinst.toolkit.driver.devices.shfqa import SHFQA
from zhinst.toolkit.driver.devices.shfsg import SHFSG
from zhinst.toolkit.driver.devices.uhfli import UHFLI
from zhinst.toolkit.driver.devices.uhfqa import UHFQA

from zhinst.toolkit.driver.devices.shfqc import SHFQC  # isort:skip

DeviceType = t.Union[BaseInstrument, HDAWG, PQSC, SHFQA, SHFSG, UHFLI, UHFQA, SHFQC]

DEVICE_CLASS_BY_MODEL = {
    "SHFQC": SHFQC,
    "SHFQA4": SHFQA,
    "SHFQA2": SHFQA,
    "SHFSG4": SHFSG,
    "SHFSG8": SHFSG,
    "HDAWG4": HDAWG,
    "HDAWG8": HDAWG,
    "PQSC": PQSC,
    "UHFQA": UHFQA,
    "UHFLI": UHFLI,
}

__all__ = [
    "BaseInstrument",
    "HDAWG",
    "PQSC",
    "SHFQA",
    "SHFSG",
    "UHFLI",
    "UHFQA",
    "SHFQC",
]



"""Zurich Instruments Toolkit (zhinst-toolkit) Command Table Module.

This module provides a :class:`CommandTable` class to create and modify ZI device
command tables.
"""

import copy
import json
import typing as t

import jsonref
import jsonschema

from zhinst.toolkit.exceptions import ValidationError

# JSON Schema validator for validating the schemes.
JSON_SCHEMA_VALIDATOR = jsonschema.Draft4Validator


def _validate_instance(instance, schema, validator=JSON_SCHEMA_VALIDATOR):
    try:
        jsonschema.validate(
            instance=instance,
            schema=schema,
            cls=validator,
        )
    except jsonschema.ValidationError as e:
        raise ValidationError(str(e)) from None


class ParentNode:
    """ParentNode of the command table.

    ParentNode can contain one or multiple arguments and child ParentNodes.
    It offers a dictionary-like object to manipulate command table properties.
    Similar to the device nodes, it supports accessing the properties by attribute.

    Args:
        schema: JSON schema of the node.
        path: Path representation of the node.
    """

    def __init__(self, schema: dict, path: t.Tuple[str, ...]):
        self._schema = schema
        self._path = path
        self._childs: t.Dict[t.Union[str, int], t.Any] = {}

    def __repr__(self) -> str:
        return "/" + "/".join(self._path)

    def is_empty(self) -> bool:
        """Check if the Node is empty and has no properties.

        Returns:
            bool: If children exists.
        """
        return not bool(self._childs)


class ParentEntry(ParentNode):
    """Parent entry of the CommandTable.

    The parent can have both properties and child properties.

    Args:
        schema: JSON schema of the node.
        path: Path representation of the node.
    """

    def __init__(self, schema: dict, path: t.Tuple[str, ...]):
        super().__init__(schema, path)
        self._attributes = {}
        self._child_props = {}
        self._properties: t.Dict[str, t.Any] = {}
        # ParentEntry does not have json list entries at this moment.
        for name, property_ in schema["properties"].items():
            if "properties" in property_:
                self._child_props[name] = property_
            else:
                self._attributes[name] = property_

    def __contains__(self, k):
        return k in self._child_props or self._attributes.get(k, None) is not None

    def __dir__(self):
        dir_info = set()
        for k in super().__dir__():
            if not k.startswith("_"):
                dir_info.add(k)
        return dir_info.union(set(list(self._schema["properties"].keys())))

    def __getattr__(self, name: str) -> t.Union["ParentEntry", t.Any]:
        if name.startswith("_"):
            return None
        try:
            return self._childs[name]
        except KeyError:
            if name in self._child_props:
                self._childs[name] = ParentEntry(
                    self._child_props[name], self._path + (name,)
                )
                return self._childs[name]
            if name in self._attributes:
                return self._properties.get(name, None)
        raise AttributeError(name)

    def __setattr__(self, name: str, value: t.Any):
        if name.startswith("_"):
            super().__setattr__(name, value)
        elif self._attributes and name in self._attributes:
            if value is None:
                self._childs.pop(name, None)
                self._properties.pop(name, None)
            else:
                _validate_instance(value, self._attributes[name])
                self._childs[name] = value
                self._properties[name] = value
        elif value is None and name in self._childs:
            self._childs.pop(name)
        else:
            raise AttributeError(name)

    def as_dict(self) -> dict:
        """Return a dictionary presentation of the table node.

        Returns:
            dict: Table node as dictionary.
        """
        result = {}
        for name, child in self._childs.items():
            if isinstance(child, (ParentEntry, HeaderEntry)):
                if not child.is_empty():
                    result[name] = child.as_dict()
                else:
                    continue
            else:
                result[name] = child
        return result

    def info(self, value: t.Optional[str] = None) -> dict:
        """Get info about the property.

        Args:
            value: Info about to specific property. Otherwise
                return info about the whole property.

        Returns:
            Info about the property.
        """
        return self._schema["properties"].get(value, None) if value else self._schema

    def clear(self) -> None:
        """Clear all properties from the object."""
        self._childs = {}
        self._properties = {}


class ListEntry(ParentNode):
    """List entry of a command table.

    Args:
        schema: JSON schema of the node.
        path: Path representation of the node.
    """

    def __init__(self, schema: dict, path: t.Tuple[str, ...]):
        super().__init__(schema, path)
        self._schema = copy.deepcopy(schema)
        self._min_length = schema["minItems"]
        self._max_length = schema["maxItems"]

        self._index_schema = schema["items"]["properties"]["index"]
        self._schema["items"]["properties"].pop("index", None)

    def __len__(self):
        return len(self._childs)

    def __getitem__(self, number: int) -> ParentEntry:
        _validate_instance(number, self._index_schema)
        try:
            return self._childs[number]
        except KeyError:
            self._childs[number] = ParentEntry(
                self._schema["items"], self._path + (str(number),)
            )
            return self._childs[number]

    def __delitem__(self, key):
        del self._childs[key]

    @property
    def range(self) -> t.Tuple[int, int]:
        """Get the range for number of minimum and maximum items in the table.

        Returns:
            Range for number of items in the table.
        """
        return (self._min_length, self._max_length)

    def as_list(self) -> t.List[dict]:
        """Return a list representation of the table.

        Returns:
            List of dictionary representation of entries in the table.
        """
        table = []
        for name, child in self._childs.items():
            if isinstance(child, ParentEntry):
                json_ = child.as_dict()
                if json_:
                    item = {"index": name}
                    item.update(json_)
                    table.append(item)
        return table


class HeaderEntry(ParentEntry):
    """Header entry of a command table.

    Args:
        schema: JSON schema of the node.
        path: Path representation of the node.
        version: JSON schema version
    """

    def __init__(self, schema: dict, path: tuple, version: t.Optional[str] = None):
        super().__init__(schema, path)
        # L1 22.08 new schema format
        if version:
            self._childs["version"] = version
        else:
            self._childs["version"] = schema["properties"]["version"]["enum"][0]

    @property
    def version(self) -> str:
        """Version of the schema."""
        return self._childs["version"]


def _derefence_json(schema: t.Union[str, dict]) -> t.Any:
    """Dereference JSON schema.

    Args:
        schema: JSON schema as a string or dictionary.

    Returns:
        Dereferenced schema.

    Raises:
        ValueError: Wrong `schema` type.
    """
    if isinstance(schema, str):
        return jsonref.loads(schema, jsonschema=True)
    if isinstance(schema, dict):
        return jsonref.loads(json.dumps(schema), jsonschema=True)
    raise ValueError(schema)


class CommandTable:
    """A representation of a ZI device command table.

    The class provides functionality to create and modify existing
    command tables.

    The CommandTable can be modified by via ``header`` and ``table`` properties.

    Args:
        json_schema: JSON Schema of the command table.

    Example:
    .. code-block:: python

        from zhinst.toolkit import CommandTable

        ct = CommandTable(json_schema)

    The ``header`` and ``table`` and then be called.

    .. code-block:: python

        ct.header.version
        # "1.1"
        ct.header.userString = "My table"

        ct.table[0].amplitude.value = 1
        ct.table[0].amplitude
        # 1

        ct.as_json()
    """

    def __init__(self, json_schema: t.Union[str, dict]):
        self._ct_schema: t.Dict = _derefence_json(json_schema)
        self._header: HeaderEntry = self._header_entry()
        self._table: ListEntry = self._table_entry()

    @property
    def header(self) -> HeaderEntry:
        """Header of the built command table."""
        return self._header

    @property
    def table(self) -> ListEntry:
        """Table entry of the built command table."""
        return self._table

    def _header_entry(self) -> HeaderEntry:
        return HeaderEntry(
            self._ct_schema["definitions"]["header"],
            ("header",),
            self._ct_schema.get("version", ""),
        )

    def _table_entry(self) -> ListEntry:
        return ListEntry(self._ct_schema["definitions"]["table"], ("table",))

    def clear(self) -> None:
        """Clear CommandTable back to its initial state."""
        self._header = self._header_entry()
        self._table = self._table_entry()

    def as_dict(self) -> dict:
        """Return a dictionary representation of the :class:`CommandTable`.

        The function formats the returner value into a schema which is
        accepted by the ZI devices which support command tables.

        The table is validated against the given schema.

        Returns:
            CommandTable as a Python dictionary.

        Raises:
            :class:`~zhinst.toolkit.exceptions.ValidateError`: The command table
                does not correspond to the given JSON schema.
        """
        result = {
            "$schema": self._ct_schema["$schema"],
            "header": self._header.as_dict(),
            "table": self._table.as_list(),
        }
        _validate_instance(result, self._ct_schema)
        return result

    def update(self, command_table: t.Union[str, dict]) -> None:
        """Update the existing instance of ``CommandTable`` with command table JSON.

        If both command tables have the same properties, the existing ones
        are overwritten by the new command table.

        Args:
            command_table: Existing command table JSON.
        """

        def json_to_dict(json_: t.Union[str, t.Dict]) -> t.Dict:
            if isinstance(json_, str):
                json_ = json.loads(json_)
            return json_  # type: ignore[return-value]

        command_table = json_to_dict(copy.deepcopy(command_table))
        _validate_instance(command_table, self._ct_schema)

        def build_nodes(path: t.Optional[ParentEntry], index: int, obj: dict):
            for k, v in obj.items():
                if isinstance(v, dict):
                    build_nodes(getattr(path, k), index, v)
                else:
                    setattr(path, k, v)

        def build_header_nodes(header: HeaderEntry, obj: dict):
            for k, v in obj.items():
                if isinstance(k, dict):
                    build_header_nodes(
                        getattr(header, k), v  # type: ignore[call-overload]
                    )
                else:
                    setattr(header, k, v)

        build_header_nodes(self._header, command_table["header"])

        for item in command_table["table"]:
            index = item.pop("index", None)
            build_nodes(self._table[index], index, item)



"""Command Table Node adaptions."""
import json
import string
import typing as t
from functools import lru_cache
from pathlib import Path

from zhinst.toolkit.command_table import CommandTable
from zhinst.toolkit.nodetree import Node, NodeTree

_CT_RESOURCE_PATH = Path(__file__).parent.parent.parent / "resources"
_CT_FILES = {
    "shfqc": _CT_RESOURCE_PATH / "ct_schema_shfsg.json",
    "shfsg": _CT_RESOURCE_PATH / "ct_schema_shfsg.json",
    "hdawg": _CT_RESOURCE_PATH / "ct_schema_hdawg.json",
}


class CommandTableNode(Node):
    """CommandTable node.

    This class implements the basic functionality of the command table allowing
    the user to load and upload their own command table.

    A dedicated class called ``CommandTable`` exists that is the preferred way
    to create a valid command table. For more information about the
    ``CommandTable`` refer to the corresponding example or the documentation
    of that class directly.

    Args:
        root: Node used for the upload of the command table
        tree: Tree (node path as tuple) of the current node
        device_type: Device type.
    """

    def __init__(
        self, root: NodeTree, tree: t.Tuple[str, ...], device_type: str
    ) -> None:
        Node.__init__(self, root, tree)
        self._device_type = device_type

    def check_status(self) -> bool:
        """Check status of the command table.

        Returns:
            Flag if a valid command table is loaded into the device.

        Raises:
            RuntimeError: If the command table upload into the device failed.
        """
        ct_status = self.status()
        if ct_status >> 3:
            raise RuntimeError(
                "Uploading of data to the command table failed "
                "due to a JSON parsing error."
            )
        return ct_status == 1

    @lru_cache()
    def load_validation_schema(self) -> t.Dict[str, t.Any]:
        """Load device command table validation schema.

        Returns:
            JSON validation schema for the device command tables.
        """
        try:
            return json.loads(self.schema())
        except KeyError:
            device_type_striped = self._device_type.lower().rstrip(string.digits)
            with open(_CT_FILES[device_type_striped], encoding="utf-8") as file_:
                return json.load(file_)

    def upload_to_device(
        self, ct: t.Union[CommandTable, str, dict], *, validate: bool = True
    ) -> None:
        """Upload command table into the device.

        The command table can either be specified through the dedicated
        ``CommandTable`` class or in a raw format, meaning a json string or json
        dict. In the case of a json string or dict the command table is
        validated by default against the schema provided by the device.

        Args:
            ct: Command table.
            validate: Flag if the command table should be validated. (Only
                applies if the command table is passed as a raw json string or
                json dict)

        Raises:
            RuntimeError: If the command table upload into the device failed.
            zhinst.toolkit.exceptions.ValidationError: Incorrect schema.
        """
        try:
            self.data(json.dumps(ct.as_dict()))  # type: ignore
        except AttributeError:
            if validate:
                ct_new = CommandTable(self.load_validation_schema())
                ct_new.update(ct)
                self.upload_to_device(ct_new)
            elif isinstance(ct, str):
                self.data(ct)
            else:
                self.data(json.dumps(ct))
        self.check_status()

    def load_from_device(self) -> CommandTable:
        """Load command table from the device.

        Returns:
            command table.
        """
        ct = CommandTable(self.load_validation_schema())
        ct.update(self.data())
        return ct



"""Module for toolkit representations of native LabOne modules."""
import typing as t

from zhinst.toolkit.driver.modules.base_module import BaseModule
from zhinst.toolkit.driver.modules.daq_module import DAQModule
from zhinst.toolkit.driver.modules.shfqa_sweeper import SHFQASweeper
from zhinst.toolkit.driver.modules.sweeper_module import SweeperModule

ModuleType = t.Union[BaseModule, DAQModule, SHFQASweeper, SweeperModule]

__all__ = ["BaseModule", "DAQModule", "SHFQASweeper", "SweeperModule"]



# Copyright (C) 2020 Zurich Instruments
#
# This software may be modified and distributed under the terms
# of the MIT license. See the LICENSE file for details.
"""Individual drivers for the devices and modules."""



"""Base Module Driver.

Natively works with all module types and provides the basic functionality like
the module specific nodetree.
"""
import logging
import time
import typing as t

from zhinst.core import ModuleBase
from zhinst.toolkit.nodetree import Node, NodeTree

logger = logging.getLogger(__name__)

if t.TYPE_CHECKING:
    from zhinst.toolkit.driver.devices import DeviceType
    from zhinst.toolkit.session import Session

ZIModule = t.TypeVar("ZIModule", bound=ModuleBase)


class BaseModule(Node):
    """Generic toolkit driver for a LabOne Modules.

    All module specific class are derived from this class.
    It exposes the nodetree and also implements common functions valid for all
    modules.
    It also can be used directly, e.g. for modules that have no special class
    in toolkit.

    Args:
        raw_module: zhinst.core module.
        session: Session to the Data Server.
    """

    def __init__(self, raw_module: ZIModule, session: "Session"):
        self._raw_module = raw_module
        self._session = session
        super().__init__(NodeTree(raw_module), tuple())
        if "device" in self.root:
            self.root.update_nodes(
                {
                    "/device": {
                        "GetParser": self._get_device,
                        "SetParser": self._set_device,
                    }
                },
                raise_for_invalid_node=False,
            )

    def __repr__(self):
        return str(f"{self._raw_module.__class__.__name__}({repr(self._session)})")

    def _get_device(self, serial: str) -> t.Union["DeviceType", str]:
        """Convert a device serial into a toolkit device object.

        Args:
            serial: Serial of the device

        Returns:
            Toolkit device object. If the serial does not
                match to a connected device the serial is returned instead.
        """
        try:
            return self._session.devices[serial]
        except (RuntimeError, KeyError):
            return serial

    @staticmethod
    def _set_device(value: t.Union["DeviceType", str]) -> str:
        """Convert a toolkit device object into a serial string.

        Args:
            serial: A toolkit device object
                (can also be a serial string directly)

        Returns:
            str: device serial
        """
        try:
            return value.serial  # type: ignore
        except AttributeError:
            return value

    def _get_node(self, node: str) -> t.Union[Node, str]:
        """Convert a raw node string into a toolkit node.

        Args:
            node (str): raw node string

        Returns:
            Toolkit node. (if the node can not be converted the raw node
                string is returned)
        """
        try:
            return self._session.raw_path_to_node(node, module=self)
        except (KeyError, RuntimeError):
            logger.error(
                f"Could not resolve {node} into a node of the sweeper module or "
                " a connected device."
            )
            return node

    @staticmethod
    def _set_node(signal: t.Union[Node, str]) -> str:
        """Convert a toolkit node into a raw node string.

        Args:
            signal: Toolkit node

        Returns:
            str: raw string node
        """
        try:
            node = signal.node_info.path  # type: ignore
        except AttributeError:
            node = signal
        return node

    def wait_done(self, *, timeout: float = 20.0, sleep_time: float = 0.5) -> None:
        """Waits until the module is finished.

        Warning: Only usable for modules that make use of the `/finished` node.

        Args:
            timeout (float): The maximum waiting time in seconds for the
                measurement (default: 20).
            sleep_time (int): Time in seconds to wait between
                requesting sweeper state. (default: 0.5)

        Raises:
            TimeoutError: The measurement is not completed before
                timeout.
        """
        start_time = time.time()
        while (
            start_time + timeout >= time.time()
            and not self._raw_module.finished()
            and self._raw_module.progress() != 1
        ):
            print(f"Progress: {(self._raw_module.progress()[0] * 100):.1f}%", end="\r")
            time.sleep(sleep_time)
        if not self._raw_module.finished() and self._raw_module.progress() != 1:
            raise TimeoutError(f"{self._raw_module.__class__.__name__} timed out.")

    def subscribe(self, signal: Node) -> None:
        """Subscribe to a node.

        The node can either be a node of this module or of a connected device.

        Args:
            signal: Node that should be subscribed to.
        """
        self._raw_module.subscribe(signal.node_info.path)

    def unsubscribe(self, signal: Node) -> None:
        """Unsubscribe from a node.

        The node can either be a node of this module or of a connected device.

        Args:
            signal: Node that should be unsubscribed from.
        """
        self._raw_module.unsubscribe(signal.node_info.path)

    def execute(self) -> None:
        """Start the module execution.

        Subscription or unsubscription is not possible until the execution is
        finished.

        .. versionadded:: 0.4.1
        """
        self._raw_module.execute()

    @property
    def raw_module(self) -> ZIModule:
        """Underlying core module."""
        return self._raw_module



from pathlib import Path
from unittest.mock import patch, PropertyMock

import pytest

from zhinst.toolkit import Session
from zhinst.toolkit.driver.devices.shfqa import SHFQA
from zhinst.toolkit.driver.devices.shfqc import SHFQC
from zhinst.toolkit.driver.devices.shfsg import SHFSG


@pytest.fixture()
def data_dir(request):
    yield Path(request.fspath).parent / "data"


@pytest.fixture()
def mock_connection():
    with patch("zhinst.toolkit.session.core.ziDAQServer", autospec=True) as connection:
        type(connection.return_value).port = PropertyMock(return_value=8004)
        type(connection.return_value).host = PropertyMock(return_value="localhost")
        type(connection.return_value).api_level = PropertyMock(return_value=6)
        yield connection


@pytest.fixture()
def nodedoc_zi_json(data_dir):
    json_path = data_dir / "nodedoc_zi.json"
    with json_path.open("r", encoding="UTF-8") as file:
        return file.read()


@pytest.fixture()
def session(nodedoc_zi_json, mock_connection):
    mock_connection.return_value.listNodesJSON.return_value = nodedoc_zi_json
    yield Session("localhost")


@pytest.fixture()
def hf2_session(data_dir, mock_connection):
    mock_connection.return_value.getString.return_value = "HF2DataServer"
    type(mock_connection.return_value).port = PropertyMock(return_value=8005)
    yield Session("localhost", hf2=True)


@pytest.fixture()
def shfqa(data_dir, mock_connection, session):

    json_path = data_dir / "nodedoc_dev1234_shfqa.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.listNodesJSON.return_value = nodes_json

    mock_connection.return_value.getString.return_value = ""
    with patch(
        "zhinst.toolkit.driver.devices.shfqa.utils.max_qubits_per_channel",
        autospec=True,
    ) as max_qubits:
        max_qubits.return_value = 16
        yield SHFQA("DEV1234", "SHFQA4", session)


@pytest.fixture()
def shfsg(data_dir, mock_connection, session):

    json_path = data_dir / "nodedoc_dev1234_shfsg.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.listNodesJSON.return_value = nodes_json

    mock_connection.return_value.getString.return_value = ""
    yield SHFSG("DEV1234", "SHFSG8", session)


@pytest.fixture()
def shfqc(data_dir, mock_connection, session):

    json_path = data_dir / "nodedoc_dev1234_shfqc.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.listNodesJSON.return_value = nodes_json

    mock_connection.return_value.getString.return_value = ""
    yield SHFQC("DEV1234", "SHFQC", session)


@pytest.fixture()
def zi_devices_json(data_dir):
    json_path = data_dir / "zi_devices.json"
    with json_path.open("r", encoding="UTF-8") as file:
        return file.read()


@pytest.fixture()
def nodedoc_dev1234_json(data_dir):
    json_path = data_dir / "nodedoc_dev1234.json"
    with json_path.open("r", encoding="UTF-8") as file:
        return file.read()


@pytest.fixture()
def mock_sweeper_daq():
    with patch(
        "zhinst.toolkit.driver.modules.shfqa_sweeper.ziDAQServer", autospec=True
    ) as connection:
        yield connection



import numpy as np
import pytest

from zhinst.toolkit.driver.modules.sweeper_module import SweeperModule


@pytest.fixture()
def sweeper_module(data_dir, mock_connection, session):
    json_path = data_dir / "nodedoc_sweeper_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.sweep.return_value.listNodesJSON.return_value = (
        nodes_json
    )
    yield SweeperModule(mock_connection.return_value.sweep(), session)


def test_repr(sweeper_module):
    assert "MagicMock(DataServerSession(localhost:8004))" in repr(sweeper_module)


def test_gridnode(sweeper_module, mock_connection, session):
    module_mock = mock_connection.return_value.sweep.return_value
    connected_devices = "dev1234"

    def get_string_side_effect(arg):
        if arg == "/zi/devices/connected":
            return connected_devices
        if arg == "/dev1234/features/devtype":
            return "Test"
        raise RuntimeError("ZIAPINotFoundException")

    mock_connection.return_value.getString.side_effect = get_string_side_effect

    # get
    module_mock.getString.return_value = "/dev1234/test/a/b"
    result = sweeper_module.gridnode()
    assert result.raw_tree == ("test", "a", "b")
    assert result.root.prefix_hide == "dev1234"

    module_mock.getString.return_value = "/dev1111/test/a/b"
    result = sweeper_module.gridnode()
    assert result == "/dev1111/test/a/b"

    # set
    sweeper_module.gridnode(session.devices["dev1234"].test.a.c)
    module_mock.set.assert_called_with("/gridnode", "/dev1234/test/a/c")

    sweeper_module.gridnode("dev123/test/hello")
    module_mock.set.assert_called_with("/gridnode", "dev123/test/hello")


def test_read(sweeper_module, mock_connection, session):
    connected_devices = "dev1234"

    def get_string_side_effect(arg):
        if arg == "/zi/devices/connected":
            return connected_devices
        if arg == "/dev1234/features/devtype":
            return "Test"
        raise RuntimeError("ZIAPINotFoundException")

    mock_connection.return_value.getString.side_effect = get_string_side_effect

    module_mock = mock_connection.return_value.sweep.return_value
    module_mock.read.return_value = {
        "/awgcontrol": np.array([0]),
        "/device": ["1234"],
        "/dev1234/demods/0/sample": [
            [
                {
                    "header": {"systemtime": np.array([1643040922025522])},
                    "x": np.array([[1, 2, 3, 4, 5]]),
                    "y": np.array([[1, 2, 4, 4, 5]]),
                    "testsignal": np.array([[9, 9, 9, 9, 9]]),
                }
            ],
        ],
    }
    result = sweeper_module.read()
    assert len(result) == 3
    result = result[session.devices["dev1234"].demods[0].sample]
    assert "systemtime" in result[0][0]["header"]
    assert np.allclose(result[0][0]["x"], np.array([[1, 2, 3, 4, 5]]))
    assert np.allclose(result[0][0]["y"], np.array([[1, 2, 4, 4, 5]]))
    assert np.allclose(result[0][0]["testsignal"], np.array([[9, 9, 9, 9, 9]]))


def test_execute(sweeper_module):
    sweeper_module.execute()
    sweeper_module.raw_module.execute.assert_called_once()



"""PQSC Instrument Driver."""

import logging
import time
from typing import List, Union

from zhinst.toolkit.driver.devices.base import BaseInstrument

logger = logging.getLogger(__name__)


class PQSC(BaseInstrument):
    """High-level driver for the Zurich Instruments PQSC."""

    def arm(self, *, deep=True, repetitions: int = None, holdoff: float = None) -> None:
        """Prepare PQSC for triggering the instruments.

        This method configures the execution engine of the PQSC and
        clears the register bank. Optionally, the *number of triggers*
        and *hold-off time* can be set when specified as keyword
        arguments. If they are not specified, they are not changed.

        Note that the PQSC is disabled at the end of the hold-off time
        after sending out the last trigger. Therefore, the hold-off time
        should be long enough such that the PQSC is still enabled when
        the feedback arrives. Otherwise, the feedback cannot be processed.

        Args:
            deep: A flag that specifies if a synchronization
                should be performed between the device and the data
                server after stopping the PQSC and clearing the
                register bank (default: True).
            repetitions: If specified, the number of triggers sent
                over ZSync ports will be set (default: None).
            holdoff: If specified, the time between repeated
                triggers sent over ZSync ports will be set. It has a
                minimum value and a granularity of 100 ns
                (default: None).

        """
        # Stop the PQSC if it is already running
        self.stop(deep=deep)
        if repetitions is not None:
            self.execution.repetitions(repetitions)
        if holdoff is not None:
            self.execution.holdoff(holdoff)
        # Clear register bank
        self.feedback.registerbank.reset(1, deep=deep)

    def run(self, *, deep: bool = True) -> None:
        """Start sending out triggers.

        This method activates the trigger generation to trigger all
        connected instruments over ZSync ports.

        Args:
            deep: A flag that specifies if a synchronization
                should be performed between the device and the data
                server after enabling the PQSC (default: True).

        """
        self.execution.enable(True, deep=deep)

    def arm_and_run(self, *, repetitions: int = None, holdoff: float = None) -> None:
        """Arm the PQSC and start sending out triggers.

        Simply combines the methods arm and run. A synchronization
        is performed between the device and the data server after
        arming and running the PQSC.

        Args:
            repetitions: If specified, the number of triggers sent
                over ZSync ports will be set (default: None).
            holdoff: If specified, the time between repeated
                triggers sent over ZSync ports will be set. It has a
                minimum value and a granularity of 100 ns
                (default: None).

        """
        self.arm(deep=True, repetitions=repetitions, holdoff=holdoff)
        self.run(deep=True)

    def stop(self, *, deep: bool = True) -> None:
        """Stop the trigger generation.

        Args:
            deep: A flag that specifies if a synchronization
                should be performed between the device and the data
                server after disabling the PQSC (default: True).

        """
        self.execution.enable(False, deep=deep)

    def wait_done(self, *, timeout: float = 10, sleep_time: float = 0.005) -> None:
        """Wait until trigger generation and feedback processing is done.

        Args:
            timeout: The maximum waiting time in seconds for the
                PQSC (default: 10).
            sleep_time: Time in seconds to wait between
                requesting PQSC state

        Raises:
            TimeoutError: If the PQSC is not done sending out all
                triggers and processing feedback before the timeout.
        """
        try:
            self.execution.enable.wait_for_state_change(
                0, timeout=timeout, sleep_time=sleep_time
            )
        except TimeoutError as error:
            raise TimeoutError("PQSC timed out.") from error

    def check_ref_clock(self, *, timeout: int = 30, sleep_time: int = 1) -> bool:
        """Check if reference clock is locked successfully.

        Args:
            timeout: Maximum time in seconds the program waits
                (default: 30).
            sleep_time: Time in seconds to wait between
                requesting the reference clock status (default: 1)

        Raises:
            ToolkitError: If the process of locking to the reference clock
                exceeds the specified timeout.
        """
        ref_clock_status = self.system.clocks.referenceclock.in_.status
        ref_clock = self.system.clocks.referenceclock.in_.source
        ref_clock_actual = self.system.clocks.referenceclock.in_.sourceactual
        try:
            ref_clock_status.wait_for_state_change(
                2, invert=True, timeout=timeout, sleep_time=sleep_time
            )
        except TimeoutError as error:
            raise TimeoutError(
                "Timeout during locking to reference clock signal"
            ) from error
        if ref_clock_status() == 0:
            return True
        if ref_clock_status() == 1 and ref_clock_actual() != ref_clock():
            ref_clock("internal", deep=True)
            logger.error(
                f"There was an error locking the device({self.serial}) "
                f"onto reference clock signal. Automatically switching to internal "
                f"reference clock. Please try again."
            )
        return False

    def check_zsync_connection(
        self,
        ports: Union[List[int], int] = 0,
        *,
        timeout: int = 30,
        sleep_time: int = 1,
    ) -> Union[List[bool], bool]:
        """Check if the ZSync connection on the given port(s) is established.

        This function checks the current status of the instrument connected to
        the given ports.

        Args:
            ports: The port numbers to check the ZSync connection for.
                It can either be a single port number given as integer or a list
                of several port numbers. (default: 0)
            timeout: Maximum time in seconds the program waits (default: 30).
            sleep_time: Time in seconds to wait between requesting the reference
                clock status (default: 1)

        Raises:
            TimeoutError: If the process of establishing a ZSync connection on
                one of the specified ports exceeds the specified timeout.
        """
        ports_list = ports if isinstance(ports, list) else [ports]

        start_time = time.time()
        status = []
        for port in ports_list:
            status.append(
                self._check_zsync_connection(
                    port,
                    timeout=max(0, timeout - (time.time() - start_time)),
                    sleep_time=sleep_time,
                )
            )
        return status if isinstance(ports, list) else status[0]

    def _check_zsync_connection(
        self, port: int, timeout: float, sleep_time: float
    ) -> bool:
        """Check if the ZSync connection on the given port is successful.

        This function checks the current status of the instrument
        connected to the given port.

        Args:
            ports: Port number to check the ZSync connection for.
            timeout: Maximum time in seconds the program waits (default: 30).
            sleep_time: Time in seconds to wait between requesting the status
                (default: 1)

        Raises:
            TimeoutError: If the process of establishing a ZSync connection the
                specified port exceeds the specified timeout.
        """
        status_node = self.zsyncs[port].connection.status
        start_time = time.time()
        try:
            status_node.wait_for_state_change(
                0, invert=True, timeout=timeout, sleep_time=sleep_time
            )
            status_node.wait_for_state_change(
                1,
                invert=True,
                timeout=max(0, timeout - (time.time() - start_time)),
                sleep_time=sleep_time,
            )
        except TimeoutError as error:
            raise TimeoutError(
                "Timeout while establishing ZSync connection to the instrument "
                f"on the port {port}."
            ) from error
        return status_node() == 2



"""High-level generic lazy node tree for the zhinst.core package."""

import fnmatch
import json
import typing as t
import warnings
from keyword import iskeyword as is_keyword

# Protocol is available in the typing module since 3.8
# Ift we only support 3.8 we should switch to t.Protocol
from typing_extensions import Protocol

from contextlib import contextmanager

from zhinst.toolkit.nodetree.helper import NodeDoc
from zhinst.toolkit.nodetree.node import Node


class Connection(Protocol):
    """Protocol class for the connection used in the nodetree.

    Every connection object used in the Nodetree is expected to have at least
    this interface in order to work with the Nodetree.
    """

    # pylint: disable=invalid-name
    def listNodesJSON(self, path: str, *args, **kwargs) -> str:
        """Returns a list of nodes with description found at the specified path."""

    def get(self, path: str, *args, **kwargs) -> object:
        """Mirrors the behavior of zhinst.core ``get`` command."""

    def getInt(self, path: str) -> int:
        """Mirrors the behavior of zhinst.core ``getInt`` command."""

    def getDouble(self, path: str) -> float:
        """Mirrors the behavior of zhinst.core ``getDouble`` command."""

    def getString(self, path: str) -> str:
        """Mirrors the behavior of zhinst.core ``getDouble`` command."""

    @t.overload
    def set(self, path: str, value: t.Any) -> None:
        """Mirrors the behavior of zhinst.core ``set`` command."""

    @t.overload
    def set(self, path: t.Union[t.List[t.Tuple[str, t.Any]]]) -> None:
        """Mirrors the behavior of zhinst.core ``set`` command."""

    def set(self, path, value=None) -> None:
        """Mirrors the behavior of zhinst.core ``set`` command."""

    def subscribe(self, path: str) -> None:
        """Mirrors the behavior of zhinst.core ``subscribe`` command."""

    def unsubscribe(self, path: str) -> None:
        """Mirrors the behavior of zhinst.core ``unsubscribe`` command."""


class Transaction:
    """Transaction Manager.

    Buffers commands (node, value pairs)

    Args:
        nodetree: Underlying Nodetree
    """

    def __init__(self, nodetree: "NodeTree"):
        self._queue: t.Optional[t.List[t.Tuple[str, t.Any]]] = None
        self._root = nodetree
        self._add_callback: t.Optional[t.Callable[[str, t.Any], None]] = None

    def start(
        self, add_callback: t.Optional[t.Callable[[str, t.Any], None]] = None
    ) -> None:
        """Start the transaction.

        Args:
            add_callback: Callback to be called when ever a node value pare has
                been added to the queue. Only valid for the started
                transaction.

        Raises:
            Runtime Error if the transaction is already in progress

        .. versionchanged:: 0.4.0

            add_callback added.
        """
        if self.in_progress():
            raise RuntimeError(
                "A transaction is already in progress. Only one transaction is "
                "possible at a time."
            )
        self._queue = []
        self._add_callback = add_callback

    def stop(self) -> None:
        """Stop the transaction."""
        self._queue = None
        self._add_callback = None

    def add(self, node: t.Union[Node, str], value: t.Any) -> None:
        """Adds a single node set command to the set transaction.

        Args:
            node: Node object or string representing the node.
            value: Value that should be set to the node.

        Raises:
            AttributeError: If no transaction is in progress.
            ValueError: If the node is passed as a string in form of a relative
                path and no prefix can be added.
        """
        try:
            self._queue.append(  # type: ignore[union-attr]
                (self._root.to_raw_path(node), value)
            )
            if self._add_callback:
                self._add_callback(*self._queue[-1])  # type: ignore[index]
        except AttributeError as exception:
            raise AttributeError("No set transaction is in progress.") from exception

    def in_progress(self) -> bool:
        """Flag if the transaction is in progress."""
        return self._queue is not None

    def result(self) -> t.Optional[t.List[t.Tuple[str, t.Any]]]:
        """Resulting transaction list.

        Result:
            List of all added node value pairs.
        """
        return self._queue


class NodeTree:
    """High-level generic lazy node tree.

    All interactions with an Zurich Instruments device or a LabOne
    module happens through manipulating nodes. The ``NodeTree`` provides a
    pythonic way for that.

    It reads all available nodes its additional information from the provided
    connection and makes them available in nested dictionary like interface. The
    interface also supports accessing the nodes by attribute.

    >>> nodetree = NodeTree(connection)
    >>> nodetree.example.nodes[8].test
        /example/nodes/8/test

    To speed up the initialization time the node tree is initialized lazy.
    Meaning the dictionary is kept as a flat dictionary and is not converted
    into a nested one. In addition the nested node objects returned by the
    ``NodeTree`` also are just simple placeholders. Only when performing
    operations on a node its validity is checked an the calls get translated to
    the correct node string. (For more information on how to manipulate nodes
    refer to :class:`zhinst.toolkit.nodetree.node.Node`).

    Examples:
        >>> nodetree = NodeTree(daq)
        >>> nodetree.dev123.demods[0].freq
        /dev123/demods/0/freq
        >>> nodetree = NodeTree(daq, prefix_hide = "dev123", list_nodes = ["/dev123/*"])
        >>> nodetree.demods[0].freq
        /dev123/demods/0/freq

    Args:
        connection: Underlying connection for the node tree. All
            operations are converted into calls to that connection.
        prefix_hide: Prefix, e.g. device id, that should be hidden in the
            nodetree. (Hidden means that users do not need to specify it and it
            will be added automatically to the nodes if necessary)
            (default = None)
        list_nodes: List of nodes that should be downloaded from the connection.
            By default all available nodes are downloaded. (default = None)
        preloaded_json: Optional preloaded node information.
            (e.g for the HF2 that does not support the `listNodesJson` function)
    """

    def __init__(
        self,
        connection: Connection,
        prefix_hide: t.Optional[str] = None,
        list_nodes: t.Optional[list] = None,
        preloaded_json: t.Optional[NodeDoc] = None,
    ):
        self._prefix_hide = prefix_hide.lower() if prefix_hide else None
        self._connection = connection
        if not list_nodes:
            list_nodes = ["*"]
        self._flat_dict: NodeDoc = {}
        if preloaded_json:
            self._flat_dict = preloaded_json
        else:
            for element in list_nodes:
                nodes_json = self.connection.listNodesJSON(element)
                self._flat_dict = {**self._flat_dict, **json.loads(nodes_json)}
        self._flat_dict = {key.lower(): value for key, value in self._flat_dict.items()}
        self._transaction = Transaction(self)
        # First Layer must be generate during initialization to calculate the
        # prefixes to keep
        self._first_layer: t.List[str] = []
        self._prefixes_keep: t.List[str] = []
        self._generate_first_layer()

    def __getattr__(self, name):
        if not name.startswith("_"):
            return Node(self, (name.lower(),))
        return None

    def __getitem__(self, name):
        name = name.lower()
        if "/" in name:
            name_list = name.split("/")
            if name_list[0]:
                return Node(self, (*name_list,))
            return Node(self, (*name_list[1:],))
        return Node(self, (name,))

    def __contains__(self, k):
        return k.lower() in self._first_layer

    def __dir__(self):
        return self._first_layer

    def __iter__(self):
        for node_raw, info in self._flat_dict.items():
            yield self.raw_path_to_node(node_raw), info

    def _generate_first_layer(self) -> None:
        """Generates the internal ``_first_layer`` list.

        The list represents the available first layer of nested nodes.

        Also create the self._prefixes_keep variable. Which is the inverse of
        the self._prefix_hide attribute.

        Raises:
            SyntaxError: If any node does not start with a leading slash.
        """
        for raw_node in self._flat_dict:
            if not raw_node.startswith("/"):
                raise SyntaxError(f"{raw_node}: Leading slash not found")
            node_split = raw_node.split("/")
            # Since we always have a leading slash we ignore the first element
            # which is empty.
            if node_split[1] == self._prefix_hide:
                if node_split[2] not in self._first_layer:
                    self._first_layer.append(node_split[2])
            else:
                if node_split[1] not in self._prefixes_keep:
                    self._prefixes_keep.append(node_split[1])
        self._first_layer.extend(self._prefixes_keep)

    def get_node_info(
        self, node: t.Union[Node, str]
    ) -> t.Dict[Node, t.Optional[t.Dict]]:
        """Get the information/data for a node.

        Unix shell-style wildcards are supported.

        Args:
            node: Node object or string representation.

        Returns:
            Node(s) information.

        Raises:
            KeyError if the node does not match an existing node.
            ValueError: If the node is passed as a string in form of a relative
                path and no prefix can be added.
        """
        key = self.to_raw_path(node)
        # resolve potential wildcards
        keys = fnmatch.filter(self._flat_dict.keys(), key)
        result = {}
        for single_key in keys:
            result[self.raw_path_to_node(single_key)] = self._flat_dict.get(single_key)
        if not result:
            raise KeyError(key)
        return result

    def update_node(
        self,
        node: t.Union[Node, str],
        updates: t.Dict[str, t.Any],
        *,
        add: bool = False,
    ) -> None:
        """Update a node in the NodeTree.

        Nodes containing wildcards will be resolved but it is not possible to
        add new nodes with a ``node`` argument containing wildcards.

        Args:
            node: Node object or string representation.
            updates: Data that will be updated (overwrites the existing values).
            add: Flag a non-existing node should be added (default = False).

        Raises:
            KeyError: If node does not exist and the ``add`` Flag is not set
            ValueError: If the node is passed as a string in form of a relative
                path and no prefix can be added.
        """
        potential_key = self.to_raw_path(node).lower()
        # resolve potential wildcards
        keys = fnmatch.filter(self._flat_dict.keys(), potential_key)
        if not keys:
            if not add:
                raise KeyError(potential_key)
            if any(wildcard in potential_key for wildcard in ["*", "?", "["]):
                # Can be implemented in the future if necessary
                raise RuntimeError(
                    f"{potential_key}: Unable to resolve wildcards when adding "
                    "new nodes."
                )
            self._flat_dict[potential_key] = updates
            first_node = potential_key.split("/")[1]
            if not self._prefix_hide == first_node:
                self._prefixes_keep.append(first_node)
                self._first_layer.append(first_node)
        else:
            for single_key in keys:
                self._flat_dict[single_key].update(updates)
        Node.node_info.fget.cache_clear()

    def update_nodes(
        self,
        update_dict: t.Dict[t.Union[Node, str], t.Dict[str, t.Any]],
        *,
        add: bool = False,
        raise_for_invalid_node: bool = True,
    ) -> None:
        """Update multiple nodes in the NodeTree.

        Similar to :func:`update_node` but for multiple elements that are
        represented as a dict.

        Args:
            update_dict: Dictionary with node as keys and entries that will be
                updated as values.
            add: Flag a non-existing node should be added (default = False).
            raise_for_invalid_node: If set to `True`, when `add` is False and the
                node(s) are invalid/nonexistent, an error is raised.

                Otherwise will issue a warning and continue adding the valid nodes.

                .. versionadded:: 0.3.4

        Raises:
            KeyError: If node does not exist and the ``add`` flag is not set
        """
        for node, updates in update_dict.items():
            try:
                self.update_node(node, updates, add=add)
            except KeyError as error:
                if raise_for_invalid_node:
                    raise
                warnings.warn(f"Tried to add nonexistent node: {error.args[0]}.")

    def raw_path_to_node(self, raw_path: str) -> Node:
        """Converts a raw node path string into a Node object.

        The function does not check if the node exists, but if the node exist
        the returned node does correspond also to that node.

        Args:
            raw_path: Raw node path (e.g. /dev1234/relative/path/to/node).

        Returns:
            The corresponding node object linked to this nodetree.
        """
        node_split = raw_path.split("/")
        # buildin keywords are escaped with a tailing underscore
        # (https://pep8.org/#descriptive-naming-styles)
        node_split = [
            element + "_" if is_keyword(element) else element for element in node_split
        ]
        # Since we always have a leading slash we ignore the first element
        # which is empty.
        if node_split[1] == self._prefix_hide:
            return Node(self, (*node_split[2:],))
        return Node(self, (*node_split[1:],))

    def to_raw_path(self, node: t.Union[Node, str]) -> str:
        """Converts a node into a raw node path string.

        The function does not check if the node exists, but if the node exist
        the returned raw node path exists in the underlying dictionary.

        Args:
            node: Node object or string representing the node.

        Returns:
            Raw node path that can be used a key in the internal dictionary.

        Raises:
            ValueError: If the node is passed as a string in form of a relative
            path and no prefix can be added.
        """
        return (
            self.node_to_raw_path(node)
            if isinstance(node, Node)
            else self.string_to_raw_path(node)
        )

    def node_to_raw_path(self, node: Node) -> str:
        """Converts a node into a raw node path string.

        The function does not check if the node exists, but if the node exist
        the returned raw node path exists in the underlying dictionary.

        Args:
            node: Node object.

        Returns:
            Raw node path that can be used a key in the internal dictionary.
        """
        if not node.raw_tree:
            return "/" + self._prefix_hide if self._prefix_hide else "/"
        # buildin keywords are escaped with a tailing underscore
        # (https://pep8.org/#descriptive-naming-styles)
        node_list = [element.rstrip("_") for element in node.raw_tree]
        if node_list[0] in self._prefixes_keep:
            string_list = "/".join(node_list)
        else:
            try:
                string_list = "/".join(
                    [self._prefix_hide] + node_list  # type: ignore[arg-type]
                )
            except TypeError:
                string_list = "/".join(node_list)
        return "/" + string_list

    def string_to_raw_path(self, node: str) -> str:
        """Converts a string representation of a node into a raw node path string.

        The function does not check if the node exists, but if the node exist
        the returned raw node path exists in the underlying dictionary.

        If the string does not represent a absolute path (leading slash) the
        :obj:`prefix_hide` will be added to the node string.

        Args:
            node: A string representation of the node.

        Returns:
            Raw node path that can be used a key in the internal dictionary.

        Raises:
            ValueError: If the node is a relative path and no prefix can be
                added.
        """
        if not node.startswith("/"):
            try:
                return "/" + self._prefix_hide + "/" + node.lower()  # type: ignore
            except TypeError as error:
                raise ValueError(
                    f"{node} is a relative path but should be a "
                    "absolute path (leading slash)"
                ) from error
        return node.lower()

    @contextmanager
    def set_transaction(self) -> t.Generator[None, None, None]:
        """Context manager for a transactional set.

        Can be used as a context in a with statement and bundles all node set
        commands into a single transaction. This reduces the network overhead
        and often increases the speed.

        Within the with block all set commands to a node will be buffered
        and grouped into a single command at the end of the context
        automatically. (All other operations, e.g. getting the value of a node,
        will not be affected)

        Warning:
            The set is always performed as deep set if called on device nodes.

        Examples:
            >>> with nodetree.set_transaction():
                    nodetree.test[0].a(1)
                    nodetree.test[1].a(2)
        """
        self._transaction.start()
        try:
            yield
            self.connection.set(self._transaction.result())  # type: ignore[arg-type]
        finally:
            self._transaction.stop()

    @property
    def transaction(self) -> Transaction:
        """Transaction manager."""
        return self._transaction

    @property
    def connection(self) -> Connection:
        """Underlying connection."""
        return self._connection

    @property
    def prefix_hide(self) -> t.Optional[str]:
        """Prefix (e.g device id), that is hidden in the nodetree.

        Hidden means that users do not need to specify it and it will be added
        automatically to the nodes if necessary.
        """
        return self._prefix_hide

    @property
    def raw_dict(self) -> dict:
        """Underlying flat dictionary with all node information."""
        return self._flat_dict



import json
from array import array
from collections import OrderedDict
from pathlib import Path
from types import GeneratorType
from unittest.mock import MagicMock, Mock
from copy import deepcopy
import pickle

import pytest
from numpy import array as nparray

from zhinst.toolkit.driver.devices import HDAWG
from zhinst.toolkit.nodetree import Node, NodeTree
from zhinst.toolkit.nodetree.connection_dict import ConnectionDict
from zhinst.toolkit.nodetree.node import NodeList


@pytest.fixture()
def data_dir(request):
    yield Path(request.fspath).parent / "data"


@pytest.fixture()
def connection(data_dir):

    json_path = data_dir / "nodedoc_dev1234_zi.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()

    connection = MagicMock()
    connection.listNodesJSON.return_value = nodes_json
    yield connection


def test_init(connection, data_dir):
    tree = NodeTree(connection)
    connection.listNodesJSON.assert_called_once_with("*")
    assert "zi" in tree
    assert "ZI" in tree
    assert "dev1234" in tree
    assert "DEV1234" in tree
    assert "zi" in dir(tree)
    assert "dev1234" in dir(tree)

    tree = NodeTree(connection, "DEV1234")
    connection.listNodesJSON.assert_called_with("*")
    assert "zi" in tree
    assert "ZI" in tree
    assert "dev1234" not in tree
    assert "DEV1234" not in tree
    assert "zi" in dir(tree)
    assert "dev1234" not in dir(tree)

    json_path = data_dir / "nodedoc_daq_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    connection_module = MagicMock()
    connection_module.listNodesJSON.return_value = nodes_json
    tree = NodeTree(connection_module)
    connection_module.listNodesJSON.assert_called_with("*")
    assert "awgcontrol" in tree

    connection_broken = MagicMock()
    connection_broken.listNodesJSON.return_value = '{\n"zi/open": {\n"Node": "/ZI/OPEN"\n},\n"zi/close": {\n"Node": "/ZI/CLOSE"\n}\n}\n'
    with pytest.raises(Exception) as e_info:
        tree = NodeTree(connection_broken)
    assert "Leading slash not found" in e_info.value.args[0]


def test_init_preloaded_json(nodedoc_dev1234_json):
    nodes_json = json.loads(nodedoc_dev1234_json)
    connection = MagicMock()  # plain mock without json info
    tree = NodeTree(connection, prefix_hide="DEV1234", preloaded_json=nodes_json)
    assert tree.prefix_hide == "dev1234"
    connection.listNodesJSON.mock_object.assert_not_called()


def test_to_raw_path(connection):
    tree = NodeTree(connection)

    with pytest.raises(ValueError):
        tree.to_raw_path("hello/world")

    assert tree.to_raw_path(Node(tree, tuple())) == "/"

    tree = NodeTree(connection, "DEV1234")
    assert tree.to_raw_path(Node(tree, tuple())) == "/dev1234"


def test_to_raw_path_capital(connection):
    tree = NodeTree(connection)
    assert tree.to_raw_path("/dev1234/demods/0/enable") == "/dev1234/demods/0/enable"
    assert tree.to_raw_path("/DEV1234/demods/0/enable") == "/dev1234/demods/0/enable"
    assert tree.to_raw_path("/dev1234/Demods/0/Enable") == "/dev1234/demods/0/enable"


def test_root_node(connection):
    tree = NodeTree(connection, "DEV1234")

    root_node = Node(tree, tuple())
    assert "demods" in root_node


def test_node_contains(connection):
    tree = NodeTree(connection, "DEV1234")

    assert "rate" in tree.demods[0]  # existing child node
    assert "invalid" not in tree.demods[0]  # non existing child node
    assert "node" not in tree.demods[0]  # property


def test_node_iter(connection):
    tree = NodeTree(connection, "DEV1234")

    childs = []
    for child, info in tree.demods[0]:
        childs.append(child)
        assert child.node_info.path.lower() == info["Node"].lower()

    assert all([child.raw_tree[-1] in tree.demods[0] for child in childs])

    childs = []
    for child, info in tree.demods["*"]:
        childs.append(child)
        assert child.node_info.path.lower() == info["Node"].lower()
    assert all([child.raw_tree[-1] in tree.demods["*"] for child in childs])


def test_node_access(connection):
    tree = NodeTree(connection, "DEV1234")

    assert tree.demods.root == tree

    # test local variable
    assert not tree._test
    assert not tree.demods._test
    # test attribute vs item
    assert tree.demods == tree["demods"]
    assert tree.demods == tree["/demods"]
    assert tree.demods.test == tree["demods/test"]
    assert tree.demods.test == tree.demods["/test"]
    assert tree.demods.test.world == tree.demods["test/world"]

    # buildin keywords are escaped with tailing underscore
    assert tree.in_.test == tree["in/test"]
    assert tree.test.and_ == tree["test/and"]

    # test invalid nodes
    assert tree.no.real.element.raw_tree == ("no", "real", "element")
    assert str(tree.no.real.element) == "/dev1234/no/real/element"
    with pytest.raises(KeyError) as e_info:
        tree.no.real.element()
    assert e_info.value.args[0] == str(tree.no.real.element)


def test_node_access_capital(connection):
    tree = NodeTree(connection, "DEV1234")
    assert tree.demods[0].enable != tree.Demods[0].Enable
    assert tree.demods[0].enable == tree["Demods/0/Enable"]


def test_node(connection):
    tree = NodeTree(connection, "DEV1234")
    # test properties of node
    assert (
        repr(tree.demods[0].rate.node_info)
        == 'NodeInfo("/dev1234/demods/0/rate",leaf-node)'
    )
    assert (
        str(tree.demods[0].rate.node_info)
        == """\
/dev1234/demods/0/rate
Defines the demodulator sampling rate, the number of samples that are sent to the host \
computer per second. A rate of about 7-10 higher as compared to the filter bandwidth \
usually provides sufficient aliasing suppression. This is also the rate of data \
received by LabOne Data Server and saved to the computer hard disk. This setting has \
no impact on the sample rate on the auxiliary outputs connectors. Note: the value \
inserted by the user may be approximated to the nearest value supported by the \
instrument.
Properties: Read, Write, Setting
Type: Double
Unit: 1/s"""
    )

    assert "Options" in str(tree.demods[0].enable.node_info)
    assert "Partial node" in str(tree.demods[0].node_info)
    assert "Contains wildcards" in str(tree.demods["*"].node_info)

    assert "description" in dir(tree.demods[0].rate.node_info)

    assert tree.demods[0].rate.node_info.path
    assert tree.demods[0].rate.node_info.description
    assert tree.demods[0].rate.node_info.type
    assert tree.demods[0].rate.node_info.unit
    assert tree.demods[0].trigger.node_info.options.get(1)
    assert tree.demods[0].trigger.node_info.is_setting
    assert not tree.demods[0].trigger.node_info.is_vector
    assert not tree.system.impedance.calib.user.data.node_info.is_setting
    assert tree.system.impedance.calib.user.data.node_info.is_vector
    assert tree.system.node_info.is_setting is None
    assert tree.system.node_info.is_vector is None

    # dir of nested node
    assert "0" in dir(tree.demods)
    assert "rate" in dir(tree.demods[0])

    # contains
    assert "Description" in tree.demods[0].rate.node_info
    assert "Test" not in tree.demods[0].rate.node_info

    # parsers
    def add_one(value):
        return value + 1

    def sub_one(value):
        return value - 1

    tree.update_node(
        "/DEV1234/test/path/0",
        {
            "SetParser": add_one,
            "GetParser": sub_one,
        },
        add=True,
    )
    assert tree.test.path[0].node_info.get_parser(5) == 4
    assert tree.test.path[0].node_info.set_parser(5) == 6

    tree.update_node(
        "/DEV1234/test/path/0",
        {"SetParser": [add_one, add_one], "GetParser": [sub_one, sub_one]},
        add=True,
    )
    assert tree.test.path[0].node_info.get_parser(5) == 3
    assert tree.test.path[0].node_info.set_parser(5) == 7

    # dynamic node
    with pytest.raises(KeyError):
        tree.demods.waves[0].node_info.unit


def test_wildcard_node(connection):
    tree = NodeTree(connection, "DEV1234")
    wildcard_node = tree.demods["*"].rate
    assert wildcard_node.node_info.path == "/dev1234/demods/*/rate"
    with pytest.raises(KeyError) as e_info:
        wildcard_node.node_info.description

    # TODO add dummy Data to mock
    wildcard_node()
    wildcard_node(1)
    with tree.set_transaction():
        wildcard_node = tree.demods["*"].rate(1)

    connection.get.return_value = None
    with pytest.raises(KeyError) as e_info:
        tree.hello["*"].rate()
    with pytest.raises(KeyError) as e_info:
        tree.hello["*"].rate(1)


def test_get_cached(connection):
    tree = NodeTree(connection, "DEV1234")

    tree.demods[0].adcselect()
    connection.getInt.assert_called_with(tree.demods[0].adcselect.node_info.path)
    tree.demods[0].rate()
    connection.getDouble.assert_called_with(tree.demods[0].rate.node_info.path)
    tree.features.serial()
    connection.getString.assert_called_with(tree.features.serial.node_info.path)

    # Complex Double
    tree.update_node("demods/0/rate", {"Type": "Complex Double"})
    tree.demods[0].rate()
    connection.getComplex.assert_called_with(tree.demods[0].rate.node_info.path)
    connection.getComplex = None
    with pytest.raises(TypeError) as e_info:
        tree.demods[0].rate()

    # ZIVectorData
    with pytest.raises(TypeError) as e_info:
        tree.system.impedance.calib.user.data()

    # ZIDemodSample
    tree.demods[0].sample()
    connection.getSample.assert_called_with(tree.demods[0].sample.node_info.path)
    connection.getSample = None
    with pytest.raises(TypeError) as e_info:
        tree.demods[0].sample()

    # ZIDIOSample
    tree.dios[0].input()
    connection.getDIO.assert_called_with(tree.dios[0].input.node_info.path)

    # ZIAdvisorWave
    tree.update_node(tree.dios[0].input, {"Type": "ZIAdvisorWave"})
    with pytest.raises(StopIteration):
        tree.dios[0].input()
    connection.get.assert_called_with(tree.dios[0].input.node_info.path, flat=True)

    # invalid type
    tree.update_node(tree.dios[0].input, {"Type": "InvalidType"})
    with pytest.raises(RuntimeError):
        tree.dios[0].input()


def test_get_deep(connection):
    tree = NodeTree(connection, "DEV1234")

    with pytest.raises(TypeError) as e_info:
        tree.demods[0].rate(deep=True)
        connection.get.assert_called_with(
            tree.demods[0].rate.node_info.path, settingsonly=False, flat=True
        )
    with pytest.raises(TypeError) as e_info:
        tree.demods[0].rate(deep=True, test=True)

    # overwrite internally used kwarg
    with pytest.raises(TypeError) as e_info:
        tree.demods[0].rate(deep=True, settingsonly=True)

    connection.get.return_value = OrderedDict()
    with pytest.raises(TypeError) as e_info:
        tree.demods[0].rate(deep=True)

    # Real data
    # double node
    connection.get.return_value = OrderedDict(
        [
            (
                "/dev1234/demods/0/rate",
                {
                    "timestamp": array("q", [1236389131550]),
                    "value": array("d", [1674.10717773]),
                },
            )
        ]
    )
    timestamp, data = tree.demods[0].rate(deep=True)
    assert timestamp == 1236389131550
    assert data == 1674.10717773
    # vector node
    connection.get.return_value = OrderedDict(
        [
            (
                "/dev1234/system/impedance/calib/user/data",
                [
                    {
                        "timestamp": 3880906635,
                        "flags": 0,
                        "vector": array("l", [123, 10, 32, 10, 125, 10]),
                    }
                ],
            )
        ]
    )
    timestamp, data = tree.demods[0].sample(deep=True)
    assert timestamp == 3880906635
    assert data == array("l", [123, 10, 32, 10, 125, 10])
    # HF2 node
    connection.get.return_value = OrderedDict(
        [("/dev1234/demods/0/rate", nparray([1674.10717773]))]
    )
    timestamp, data = tree.demods[0].sample(deep=True)
    assert data == 1674.10717773
    assert timestamp == None


def test_get_wildcard(connection):
    tree = NodeTree(connection, "DEV1234")

    connection.get.return_value = OrderedDict(
        [
            (
                "/dev1234/demods/0/impedance",
                {
                    "timestamp": array("q", [3880906635]),
                    "value": array("l", [123]),
                },
            )
        ]
    )
    result = tree.demods()
    assert result[tree.demods[0].impedance] == 123
    connection.get.assert_called_with("/dev1234/demods", settingsonly=False, flat=True)

    # deep includes the timestamp
    result = tree.demods(deep=True)
    assert result[tree.demods[0].impedance] == (3880906635, 123)

    # nested dict
    result = tree.demods(flat=False)
    assert isinstance(result, OrderedDict)

    # nested dict
    result = tree.demods["*"].impedance(flat=False)
    assert isinstance(result, OrderedDict)

    # invalid node
    with pytest.raises(KeyError) as e_info:
        tree.invalid()

    # HF2 support
    connection.get.return_value = OrderedDict(
        [("/dev1234/demods/0/impedance", nparray([125]))]
    )
    result = tree.demods()
    assert result[tree.demods[0].impedance] == 125
    connection.get.assert_called_with("/dev1234/demods", settingsonly=False, flat=True)

    result = tree.demods(deep=True)
    assert result[tree.demods[0].impedance] == (None, 125)

    connection.get.return_value = OrderedDict(
        [
            (
                "/dev1234/demods/0/impedance",
                {
                    "timestamp": array("q", [3880906635]),
                    "x": array("l", [123]),
                    "y": array("l", [123]),
                },
            )
        ]
    )
    result = tree.demods()
    assert (
        result[tree.demods[0].impedance]
        == connection.get.return_value["/dev1234/demods/0/impedance"]
    )
    connection.get.assert_called_with("/dev1234/demods", settingsonly=False, flat=True)


def test_module_get_wildcard(connection):
    tree = NodeTree(connection, "DEV1234")

    return_value = OrderedDict(
        [
            (
                "/dev1234/demods/0/impedance",
                {
                    "timestamp": array("q", [3880906635]),
                    "value": array("l", [123]),
                },
            )
        ]
    )
    connection.get.side_effect = [TypeError(), return_value]
    assert tree.demods()[tree.demods[0].impedance] == 123

    connection.get.side_effect = [TypeError(), RuntimeError(), return_value]
    assert tree.demods()[tree.demods[0].impedance] == 123

    connection.get.side_effect = [RuntimeError(), return_value]
    assert tree.demods()[tree.demods[0].impedance] == 123


def test_set(connection):
    tree = NodeTree(connection, "DEV1234")
    tree.demods[0].rate(22.0)
    connection.set.assert_called_with(tree.demods[0].rate.node_info.path, 22.0)
    tree.demods[0].rate(22.0, test=True)
    connection.set.assert_called_with(
        tree.demods[0].rate.node_info.path, 22.0, test=True
    )

    # sync set of nodes is only dependend on the input parameter not the
    # type of the node iself
    tree.demods[0].rate(22.0, deep=True)
    connection.syncSetDouble.assert_called_with(
        tree.demods[0].rate.node_info.path, 22.0
    )
    tree.demods[0].rate(0, deep=True)
    connection.syncSetInt.assert_called_once_with(tree.demods[0].rate.node_info.path, 0)
    tree.demods[0].rate("test", deep=True)
    connection.syncSetString.assert_called_once_with(
        tree.demods[0].rate.node_info.path, "test"
    )
    with pytest.raises(Exception) as e_info:
        tree.demods[0].rate(complex(2, -3), deep=True)

    connection.syncSetString = None
    with pytest.raises(TypeError):
        tree.demods[0].rate("test", deep=True)

    with pytest.raises(KeyError):
        tree.demods.test("test")


def test_runtimerror_set_set_vector(connection):
    # tree.system.impedance.calib.user.data does work with set, but for
    # testing purposes call it
    tree = NodeTree(connection, "DEV1234")
    connection.set = Mock(side_effect=RuntimeError)
    tree.system.impedance.calib.user.data(b"Test Binary Data")
    connection.setVector.assert_called_with(
        tree.system.impedance.calib.user.data.node_info.path, b"Test Binary Data"
    )


def test_transactional_set(connection):
    tree = NodeTree(connection, "DEV1234")
    with tree.set_transaction():
        tree.demods[0].rate(22.0)
    connection.set.assert_called_once_with([("/dev1234/demods/0/rate", 22.0)])
    with tree.set_transaction():
        tree.demods[0].rate(22.0)
        tree.demods[0].phaseshift(1)
    connection.set.assert_called_with(
        [("/dev1234/demods/0/rate", 22.0), ("/dev1234/demods/0/phaseshift", 1)]
    )
    with tree.set_transaction():
        tree.demods[0].rate(22.0)
        tree.transaction.add("demods/0/phaseshift", 2)
    connection.set.assert_called_with(
        [("/dev1234/demods/0/rate", 22.0), ("/dev1234/demods/0/phaseshift", 2)]
    )

    with tree.set_transaction():
        tree.system.impedance.calib.user.data(b"Test Binary Data")
    connection.set.assert_called_with(
        [("/dev1234/system/impedance/calib/user/data", b"Test Binary Data")]
    )

    with pytest.raises(AttributeError) as e_info:
        tree.transaction.add("demods/0/phaseshift", 2)

    with pytest.raises(RuntimeError) as e_info:
        with tree.set_transaction():
            with tree.set_transaction():
                pass


def test_get_node_info(connection):
    tree = NodeTree(connection, "DEV1234")
    # raw get
    assert tree.get_node_info("demods/0/rate") == tree.get_node_info(
        tree.demods[0].rate
    )
    assert tree.get_node_info("/dev1234/demods/0/rate") == tree.get_node_info(
        tree.demods[0].rate
    )
    with pytest.raises(KeyError) as e_info:
        tree.get_node_info("zi/config/open")
    assert tree.get_node_info("/zi/config/open") == tree.get_node_info(
        tree.zi.config.open
    )
    with pytest.raises(KeyError) as e_info:
        tree.get_node_info("/demods/0/rate")
    assert e_info.value.args[0] == "/demods/0/rate"
    with pytest.raises(KeyError):
        tree.get_node_info("dev1234/demods/0/rate")
    with pytest.raises(KeyError) as e_info:
        tree.get_node_info("no/real/element")
    assert e_info.value.args[0] == "/dev1234/no/real/element"

    # raw get wildcards
    assert tree.get_node_info("demods/*/rate") == tree.get_node_info("demods/?/rate")
    assert len(tree.get_node_info("demods/0/*")) > 1
    with pytest.raises(KeyError) as e_info:
        tree.get_node_info("demods/0/real/*")
    assert e_info.value.args[0] == "/dev1234/demods/0/real/*"
    with pytest.raises(KeyError) as e_info:
        tree.get_node_info("demod/0/*")
    assert e_info.value.args[0] == "/dev1234/demod/0/*"


def test_update_node(connection):
    tree = NodeTree(connection, "DEV1234")

    tree.update_node("demods/0/rate", {"Unit": "test"})
    assert tree.demods[0].rate.node_info.unit == "test"
    tree.update_node("demods/0/rate", {"Test": "test"})
    assert (
        next(iter(tree.get_node_info("demods/0/rate").values())).get("Test", None)
        == "test"
    )
    tree.update_node("demods/0/rate", {"Unit": "test2", "Test2": True})
    assert tree.demods[0].rate.node_info.unit == "test2"
    assert next(iter(tree.get_node_info("demods/0/rate").values())).get("Test2", None)

    # new Node
    with pytest.raises(KeyError) as e_info:
        tree.update_node("/test", {"Node": "test"})
    assert e_info.value.args[0] == "/test"

    assert str(tree.test) == "/dev1234/test"
    tree.update_node("/test", {"Node": "test"}, add=True)
    tree.get_node_info(tree.test)
    assert tree.test.node_info["Node"] == "test"

    assert str(tree.demods[0].test) == "/dev1234/demods/0/test"
    tree.update_node("demods/0/test", {"Node": "test"}, add=True)
    assert str(tree.demods[0].test) == "test"

    tree.update_node("demods/0/rate", {"Properties": "Write"})
    with pytest.raises(Exception) as e_info:
        tree.demods[0].rate()
    tree.update_node("demods/0/rate", {"Properties": "Read"})
    with pytest.raises(Exception) as e_info:
        tree.demods[0].rate(12)

    with pytest.raises(RuntimeError) as e_info:
        tree.update_node("hello/*/world", {}, add=True)


def test_update_nodes(connection):
    tree = NodeTree(connection, "DEV1234")

    tree.update_nodes(
        {"demods/0/rate": {"Unit": "test"}, "demods/0/order": {"Unit": "test2"}}
    )
    assert tree.demods[0].rate.node_info.unit == "test"
    assert tree.demods[0].order.node_info.unit == "test2"

    with pytest.raises(KeyError) as e_info:
        tree.update_nodes(
            {"demods/0/rate": {"Unit": "test3"}, "test": {"Node": "test"}}
        )
    assert tree.demods[0].rate.node_info.unit == "test3"
    tree.update_nodes(
        {"demods/0/rate": {"Unit": "test3"}, "test": {"Node": "test4"}}, add=True
    )
    assert tree.test.node_info.path == "test4"
    assert tree.test.is_valid() is True

    # Test for not raising any errors
    tree.update_nodes(
        {"312/123": {"Unit": "test5"}, "testNOtexists": {"Node": "test5"}},
        add=False,
        raise_for_invalid_node=False,
    )
    assert tree.testNOtexists.is_valid() is False


def test_options(connection):
    tree = NodeTree(connection, "DEV1234")

    tree.demods[0].trigger("continuous")
    connection.set.assert_called_with(
        tree.demods[0].trigger.node_info.path, "continuous"
    )

    tree.demods[0].trigger("continuou")
    connection.set.assert_called_with(
        tree.demods[0].trigger.node_info.path, "continuou"
    )

    connection.getInt.return_value = 0
    assert tree.demods[0].trigger() == tree.demods[0].trigger.node_info.enum.continuous

    connection.getInt.return_value = 12345678
    assert tree.demods[0].trigger() == 12345678


def test_nameless_options(connection):
    tree = NodeTree(connection, "DEV1234")

    options = tree.demods[0].order.node_info.options
    assert len(options) == 8
    assert options[1].enum == ""
    assert options[1].description == "1st order filter 6 dB/oct"

    assert tree.demods[0].order.node_info.enum == None


def test_parser(connection):
    tree = NodeTree(connection, "DEV1234")

    tree.update_node(
        "demods/0/rate",
        {
            "GetParser": lambda value: value + 10,
            "SetParser": lambda value: value + 10,
        },
    )
    connection.getDouble.return_value = 0
    assert tree.demods[0].rate() == 10
    tree.demods[0].rate(0)
    connection.set.assert_called_with(tree.demods[0].rate.node_info.path, 10)


def test_get_as_event(connection):
    tree = NodeTree(connection, "DEV1234")

    tree.demods[0].sample.get_as_event()
    connection.getAsEvent.assert_called_once_with(tree.demods[0].sample.node_info.path)

    connection.getAsEvent.side_effect = [RuntimeError(), None, None]
    tree.demods["[1,2]"].sample.get_as_event()
    connection.getAsEvent.assert_any_call(tree.demods[1].sample.node_info.path)
    connection.getAsEvent.assert_any_call(tree.demods[2].sample.node_info.path)

    connection.getAsEvent.side_effect = RuntimeError()
    with pytest.raises(KeyError):
        tree.demods["*"].test.get_as_event()


def test_child_nodes(connection):
    tree = NodeTree(connection, "DEV1234")

    tree.demods[0].child_nodes()
    connection.listNodes.assert_not_called()
    generator = tree.demods[0].child_nodes()
    assert isinstance(generator, GeneratorType)
    connection.listNodes.assert_not_called()
    list(generator)
    connection.listNodes.assert_called_once_with(
        tree.demods[0].node_info.path,
        recursive=False,
        leavesonly=False,
        settingsonly=False,
        streamingonly=False,
        subscribedonly=False,
        basechannelonly=False,
        excludestreaming=False,
        excludevectors=False,
    )
    list(
        tree.demods[0].sample.child_nodes(
            recursive=True,
            leavesonly=False,
            settingsonly=True,
            streamingonly=False,
            subscribedonly=True,
            basechannelonly=False,
            excludestreaming=True,
            excludevectors=False,
        )
    )
    connection.listNodes.assert_called_with(
        tree.demods[0].sample.node_info.path,
        recursive=True,
        leavesonly=False,
        settingsonly=True,
        streamingonly=False,
        subscribedonly=True,
        basechannelonly=False,
        excludestreaming=True,
        excludevectors=False,
    )
    connection.listNodes.return_value = [
        tree.demods[0].sample.node_info.path,
        tree.demods[0].rate.node_info.path,
    ]
    result = list(
        tree.demods[0].child_nodes(
            recursive=True,
            leavesonly=True,
            settingsonly=True,
            streamingonly=True,
            subscribedonly=True,
            basechannelonly=True,
            excludestreaming=True,
            excludevectors=True,
        )
    )
    connection.listNodes.assert_called_with(
        tree.demods[0].node_info.path,
        recursive=True,
        leavesonly=True,
        settingsonly=True,
        streamingonly=True,
        subscribedonly=True,
        basechannelonly=True,
        excludestreaming=True,
        excludevectors=True,
    )
    assert result == [tree.demods[0].sample, tree.demods[0].rate]

    # wildcard nodes
    connection.listNodes.side_effect = RuntimeError("Weird fail")
    with pytest.raises(RuntimeError) as e_info:
        result = list(tree.demods["*"].child_nodes())
    assert e_info.value.args[0] == "Weird fail"
    connection.listNodes.side_effect = RuntimeError("Path format invalid")
    with pytest.raises(RuntimeError) as e_info:
        result = list(tree.demods["[0,1]"].child_nodes())
    assert "full_wildcard" in e_info.value.args[0]

    result = list(tree.demods["*"].child_nodes(full_wildcard=True))
    assert result


def test_subscribe(connection):
    tree = NodeTree(connection, "DEV1234")

    tree.demods[0].sample.subscribe()
    connection.subscribe.assert_called_once_with(tree.demods[0].sample.node_info.path)

    tree.demods[0].sample.unsubscribe()
    connection.unsubscribe.assert_called_once_with(tree.demods[0].sample.node_info.path)

    connection.subscribe.side_effect = RuntimeError()
    connection.unsubscribe.side_effect = RuntimeError()
    with pytest.raises(KeyError):
        tree.demods["*"].test.subscribe()
    with pytest.raises(KeyError):
        tree.demods["*"].test.unsubscribe()


def test_wait_for_state_change(connection):
    tree = NodeTree(connection, "DEV1234")

    connection.getInt.side_effect = [1] * 3 + [2] * 2
    tree.demods[0].trigger.wait_for_state_change(2)
    connection.getInt.side_effect = None

    connection.getInt.return_value = 2
    tree.demods[0].trigger.wait_for_state_change(2)

    connection.getInt.return_value = 1
    with pytest.raises(TimeoutError) as e:
        tree.demods[0].trigger.wait_for_state_change(2, timeout=0.5)
    assert (
        str(e.value) == "/dev1234/demods/0/trigger did not change to the "
        "expected value within 0.5s. 2 != 1"
    )

    with pytest.raises(TimeoutError) as e:
        tree.demods[0].trigger.wait_for_state_change(1, invert=True, timeout=0.5)
    assert (
        str(e.value) == "/dev1234/demods/0/trigger did not change from the "
        "expected value 1 within 0.5s."
    )

    connection.getInt.side_effect = [1] * 3 + [2] * 8
    tree.demods["*"].trigger.wait_for_state_change(2)

    with pytest.raises(KeyError) as e_info:
        tree.hello.wait_for_state_change(1)

    with pytest.raises(KeyError) as e_info:
        tree.hello["*"].test.wait_for_state_change(1)


def test_nodetree_iterator(connection):
    tree = NodeTree(connection, "DEV1234")

    counter = 0
    for node, info in tree:
        assert node.node_info.path.lower() == info["Node"].lower()
        counter = counter + 1
    assert counter == len(tree._flat_dict)


def test_bool(connection):
    tree = NodeTree(connection, "DEV1234")
    assert bool(tree.demods[0].freq)
    assert not bool(tree.demods[0].req)


def test_len(connection):
    tree = NodeTree(connection, "DEV1234")
    assert len(tree.demods) == 4
    with pytest.raises(TypeError):
        len(tree.demods[0])


def test_hash_node(connection):
    tree = NodeTree(connection, "DEV1234")

    test_dict = {tree.demods[0].rate: 0, tree.demods[0].trigger: 1}

    assert test_dict[tree.demods[0].rate] == 0
    assert test_dict[tree.demods[0].trigger] == 1


def test_connection_dict(data_dir):
    data = {"/car/seat": 4, "/car/color": "blue", "/street/length": 110.4}
    json_path = data_dir / "nodedoc_fake.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = json.loads(file.read())
    connection = ConnectionDict(data, nodes_json)

    tree = NodeTree(connection)
    assert tree.car.seat() == 4
    assert tree.car.color() == "blue"
    assert tree.street.length() == 110.4

    tree.car.seat(5)
    assert tree.car.seat() == 5

    tree.street.length(1)
    assert tree.street.length() == 1.0
    assert tree.street.length(deep=True) == (None, 1.0)

    tree.car["*"](1)
    assert tree.car.seat() == 1
    assert tree.car.color() == "1"

    # subscribe and unsubscribe is not supported
    with pytest.raises(RuntimeError) as e_info:
        tree.car.seat.subscribe()
    with pytest.raises(RuntimeError) as e_info:
        tree.car.seat.unsubscribe()

    tree = NodeTree(connection, list_nodes=["/car/*"])
    assert tree.car.seat() == 1
    assert tree.car.color() == "1"
    with pytest.raises(KeyError):
        tree.street.length()

    connection.setVector("/car/seat", 5)
    assert tree.car.seat() == 5


@pytest.fixture()
def hdawg(data_dir, mock_connection, session):

    json_path = data_dir / "nodedoc_dev1234_hdawg.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.listNodesJSON.return_value = nodes_json

    mock_connection.return_value.listNodes.return_value = [
        "/dev1234/awgs/0",
        "/dev1234/awgs/1",
        "/dev1234/awgs/2",
        "/dev1234/awgs/3",
    ]
    mock_connection.return_value.getString.return_value = ""

    yield HDAWG("DEV1234", "HDAWG4", session)


def test_nodelist_is_node(connection, hdawg):
    nt = NodeTree(connection, "DEV1234")
    bar = NodeList([hdawg], nt, ("foobar",))
    assert bar[0] == hdawg
    assert bar == Node(nt, ("foobar",))


class TestWildCardResult:
    @pytest.fixture()
    def node_tree(self, connection):
        tree = NodeTree(connection, "DEV1234")
        connection.get.return_value = OrderedDict(
            [
                (
                    "/dev1234/demods/0/rate",
                    {
                        "timestamp": array("q", [123]),
                        "value": array("d", [2]),
                    },
                ),
                (
                    "/dev1234/demods/1/rate",
                    {
                        "timestamp": array("q", [1234]),
                        "value": array("d", [3]),
                    },
                ),
            ]
        )
        return tree

    @pytest.fixture
    def wildcard_call_no_deep(self, node_tree):
        return node_tree.demods["*"].rate(deep=False)

    def test_call_no_deep(self, wildcard_call_no_deep):
        assert wildcard_call_no_deep == {
            "/dev1234/demods/0/rate": 2,
            "/dev1234/demods/1/rate": 3,
        }

    def test_repr(self, wildcard_call_no_deep):
        assert repr(wildcard_call_no_deep) == str(
            {"/dev1234/demods/0/rate": 2.0, "/dev1234/demods/1/rate": 3.0}
        )  # repr() displays here as float, but in Notebooks and prints it is int

    def test_getitem(self, wildcard_call_no_deep, node_tree):
        assert wildcard_call_no_deep[node_tree.demods[1].rate] == 3

    def test_assert_in(self, wildcard_call_no_deep, node_tree):
        assert node_tree.demods[1].rate in wildcard_call_no_deep

    def test_to_dict(self, wildcard_call_no_deep):
        assert wildcard_call_no_deep.to_dict() == {
            "/dev1234/demods/0/rate": 2,
            "/dev1234/demods/1/rate": 3,
        }

    def test_len(self, wildcard_call_no_deep):
        assert len(wildcard_call_no_deep) == 2

    def test_iter(self, wildcard_call_no_deep):
        assert list(iter(wildcard_call_no_deep)) == [
            "/dev1234/demods/0/rate",
            "/dev1234/demods/1/rate",
        ]

    def test_deepcopy(self, wildcard_call_no_deep):
        assert deepcopy(wildcard_call_no_deep) == {
            "/dev1234/demods/0/rate": 2,
            "/dev1234/demods/1/rate": 3,
        }

    def test_deep_wildcard(self, node_tree):
        data_wc = node_tree.demods["*"].rate(deep=True)
        assert data_wc == {
            "/dev1234/demods/0/rate": (123, 2),
            "/dev1234/demods/1/rate": (1234, 3),
        }

    def test_pickle_enum(self, node_tree, connection):
        connection.getInt.return_value = 0
        enum_value = node_tree.demods[0].trigger()
        new_enum_value = pickle.loads(pickle.dumps(enum_value))
        assert new_enum_value == enum_value
        for old, new in zip(type(enum_value), type(new_enum_value)):
            assert old == new



import pytest

from zhinst.toolkit.driver.devices.hdawg import AWG, HDAWG
from zhinst.toolkit.nodetree import Node


@pytest.fixture()
def hdawg(data_dir, mock_connection, session):

    json_path = data_dir / "nodedoc_dev1234_hdawg.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.listNodesJSON.return_value = nodes_json

    mock_connection.return_value.listNodes.return_value = [
        "/dev1234/awgs/0",
        "/dev1234/awgs/1",
        "/dev1234/awgs/2",
        "/dev1234/awgs/3",
    ]
    mock_connection.return_value.getString.return_value = ""

    yield HDAWG("DEV1234", "HDAWG4", session)


def test_repr(hdawg):
    assert repr(hdawg) == "HDAWG(HDAWG4,DEV1234)"


def test_enable_qccs_mode(mock_connection, hdawg):
    hdawg.enable_qccs_mode()
    mock_connection.return_value.set.assert_called_with(
        [
            ("/dev1234/system/clocks/referenceclock/source", "zsync"),
            ("/dev1234/dios/0/interface", 0),
            ("/dev1234/dios/0/mode", "qccs"),
            ("/dev1234/dios/0/drive", 12),
            ("/dev1234/awgs/0/dio/strobe/slope", "off"),
            ("/dev1234/awgs/1/dio/strobe/slope", "off"),
            ("/dev1234/awgs/2/dio/strobe/slope", "off"),
            ("/dev1234/awgs/3/dio/strobe/slope", "off"),
            ("/dev1234/awgs/0/dio/valid/polarity", "none"),
            ("/dev1234/awgs/1/dio/valid/polarity", "none"),
            ("/dev1234/awgs/2/dio/valid/polarity", "none"),
            ("/dev1234/awgs/3/dio/valid/polarity", "none"),
        ]
    )
    mock_connection.reset_mock()
    with hdawg.set_transaction():
        hdawg.enable_qccs_mode()
    mock_connection.return_value.set.assert_called_with(
        [
            ("/dev1234/system/clocks/referenceclock/source", "zsync"),
            ("/dev1234/dios/0/interface", 0),
            ("/dev1234/dios/0/mode", "qccs"),
            ("/dev1234/dios/0/drive", 12),
            ("/dev1234/awgs/0/dio/strobe/slope", "off"),
            ("/dev1234/awgs/1/dio/strobe/slope", "off"),
            ("/dev1234/awgs/2/dio/strobe/slope", "off"),
            ("/dev1234/awgs/3/dio/strobe/slope", "off"),
            ("/dev1234/awgs/0/dio/valid/polarity", "none"),
            ("/dev1234/awgs/1/dio/valid/polarity", "none"),
            ("/dev1234/awgs/2/dio/valid/polarity", "none"),
            ("/dev1234/awgs/3/dio/valid/polarity", "none"),
        ]
    )


def test_awg(data_dir, mock_connection, hdawg):
    json_path = data_dir / "nodedoc_awg_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.awgModule.return_value.listNodesJSON.return_value = (
        nodes_json
    )
    assert len(hdawg.awgs) == 4
    assert isinstance(hdawg.awgs[0], AWG)
    # Wildcards nodes will be converted into normal Nodes
    assert not isinstance(hdawg.awgs["*"], AWG)
    assert isinstance(hdawg.awgs["*"], Node)



"""SHFQA Instrument Driver."""

import logging
import typing as t
import warnings

import zhinst.utils.shfqa as utils

from zhinst.toolkit.driver.devices.base import BaseInstrument
from zhinst.toolkit.driver.nodes.awg import AWG
from zhinst.toolkit.driver.nodes.readout import Readout
from zhinst.toolkit.driver.nodes.shfqa_scope import SHFScope
from zhinst.toolkit.driver.nodes.spectroscopy import Spectroscopy
from zhinst.toolkit.interface import SHFQAChannelMode
from zhinst.toolkit.nodetree import Node, NodeTree
from zhinst.toolkit.nodetree.helper import (
    lazy_property,
    create_or_append_set_transaction,
)
from zhinst.toolkit.nodetree.node import NodeList
from zhinst.toolkit.waveform import Waveforms

logger = logging.getLogger(__name__)

if t.TYPE_CHECKING:
    from zhinst.toolkit.session import Session


class Generator(AWG):
    """Generator node.

    Implements basic functionality of the generator allowing the user to write
    and upload their *'.seqC'* code.

    In contrast to other AWG Sequencers, e.g. from the HDAWG, SHFSG
    it does not provide writing access to the Waveform Memories
    and hence does not come with predefined waveforms such as `gauss`
    or `ones`. Therefore, all waveforms need to be defined in Python
    and uploaded to the device using `upload_waveforms` method.

    Args:
        root: Root of the nodetree
        tree: Tree (node path as tuple) of the current node
        daq_server: Instance of the ziDAQServer
        serial: Serial of the device.
        index: Index of the corresponding awg channel
        max_qubits_per_channel: Max qubits per channel
    """

    def __init__(
        self,
        root: NodeTree,
        tree: tuple,
        serial: str,
        index: int,
        max_qubits_per_channel: int,
        device_type: str,
        device_options: str,
    ):
        super().__init__(root, tree, serial, index, device_type, device_options)
        self._max_qubits_per_channel = max_qubits_per_channel

    def write_to_waveform_memory(
        self, pulses: t.Union[Waveforms, dict], *, clear_existing: bool = True
    ) -> None:
        """Writes pulses to the waveform memory.

        Args:
            pulses: Waveforms that should be uploaded.
            clear_existing: Flag whether to clear the waveform memory before the
                present upload. (default = True)
        """
        if (
            len(pulses.keys()) > 0
            and max(pulses.keys()) >= self._max_qubits_per_channel
        ):
            raise RuntimeError(
                f"The device only has {self._max_qubits_per_channel} qubits per channel"
                f", but {max(pulses.keys())} where specified"
            )
        with create_or_append_set_transaction(self._root):
            if clear_existing:
                self.clearwave(1)
            if isinstance(pulses, Waveforms):
                for slot in pulses.keys():
                    self.waveforms[slot].wave(
                        pulses.get_raw_vector(slot, complex_output=True)
                    )
            else:
                for slot, waveform in pulses.items():
                    self.waveforms[slot].wave(waveform)

    def read_from_waveform_memory(self, slots: t.List[int] = None) -> Waveforms:
        """Read pulses from the waveform memory.

        Args:
            slots: List of waveform indexes to read from the device. If not
                specified all assigned waveforms will be downloaded.

        Returns:
            Mutable mapping of the downloaded waveforms.
        """
        nodes = []
        if slots is not None:
            for slot in slots:
                nodes.append(self.waveforms[slot].wave.node_info.path)
        else:
            nodes.append(self.waveforms["*"].wave.node_info.path)
        nodes_str = ",".join(nodes)
        waveforms_raw = self._daq_server.get(nodes_str, settingsonly=False, flat=True)
        waveforms = Waveforms()
        for slot, waveform in enumerate(waveforms_raw.values()):
            waveforms[slot] = waveform[0]["vector"]
        return waveforms

    def configure_sequencer_triggering(
        self, *, aux_trigger: str, play_pulse_delay: float = 0.0
    ) -> None:
        """Configure the sequencer triggering.

        Args:
            aux_trigger: Alias for the trigger source used in the sequencer.
                For the list of available values, use `available_aux_trigger_inputs`
            play_pulse_delay: Delay in seconds before the start of waveform playback.
        """
        # Only Digital Trigger 1
        utils.configure_sequencer_triggering(
            self._daq_server,
            self._serial,
            self._index,
            aux_trigger=aux_trigger,
            play_pulse_delay=play_pulse_delay,
        )

    @property
    def available_aux_trigger_inputs(self) -> t.List[str]:
        """List of available aux trigger sources for the generator."""
        return [
            option.enum
            for option in self.auxtriggers[0].channel.node_info.options.values()
        ]


class QAChannel(Node):
    """Quantum Analyzer Channel for the SHFQA.

    :class:`QAChannel` implements basic functionality to configure QAChannel
    settings of the :class:`SHFQA` instrument.
    Besides the :class:`Generator`, :class:`Readout` and :class:`Sweeper`
    modules it also provides an easy access to commonly used `QAChannel` parameters.

    Args:
        device: SHFQA device object.
        session: Underlying session.
        tree: Node tree (node path as tuple) of the corresponding node.
    """

    def __init__(
        self,
        device: "SHFQA",
        session: "Session",
        tree: t.Tuple[str, ...],
    ):
        super().__init__(device.root, tree)
        self._index = int(tree[-1])
        self._device = device
        self._serial = device.serial
        self._session = session

    def configure_channel(
        self,
        *,
        input_range: int,
        output_range: int,
        center_frequency: float,
        mode: SHFQAChannelMode,
    ) -> None:
        """Configures the RF input and output of a specified channel.

        Args:
            input_range: Maximal range of the signal input power in dBm
            output_range: Maximal range of the signal output power in dBm
            center_frequency: Center frequency of the analysis band [Hz]
            mode: Select between spectroscopy and readout mode.
        """
        utils.configure_channel(
            self._session.daq_server,
            self._serial,
            self._index,
            input_range=input_range,
            output_range=output_range,
            center_frequency=center_frequency,
            mode=mode.value,
        )

    @lazy_property
    def generator(self) -> Generator:
        """Generator."""
        return Generator(
            self._root,
            self._tree + ("generator",),
            self._device.serial,
            self._index,
            self._device.max_qubits_per_channel,
            self._device.device_type,
            self._device.device_options,
        )

    @lazy_property
    def readout(self) -> Readout:
        """Readout."""
        return Readout(
            self._root,
            self._tree + ("readout",),
            self._device.serial,
            self._index,
            self._device.max_qubits_per_channel,
        )

    @lazy_property
    def spectroscopy(self) -> Spectroscopy:
        """Spectroscopy."""
        return Spectroscopy(
            self._root,
            self._tree + ("spectroscopy",),
            self._device.serial,
            self._index,
        )


class SHFQA(BaseInstrument):
    """High-level driver for the Zurich Instruments SHFQA."""

    def factory_reset(self, *, deep: bool = True) -> None:
        """Load the factory default settings.

        Args:
            deep: A flag that specifies if a synchronization
                should be performed between the device and the data
                server after loading the factory preset (default: True).
        """
        warnings.warn("Factory preset is not yet supported for SHFQA.", RuntimeWarning)
        logger.warning("Factory preset is not yet supported in SHFQA.")

    def start_continuous_sw_trigger(
        self, *, num_triggers: int, wait_time: float
    ) -> None:
        """Issues a specified number of software triggers.

        Issues a specified number of software triggers with a certain wait time
        in between. The function guarantees reception and proper processing of
        all triggers by the device, but the time between triggers is
        non-deterministic by nature of software triggering. Only use this
        function for prototyping and/or cases without strong timing requirements.

        Args:
            num_triggers: Number of triggers to be issued
            wait_time: Time between triggers in seconds
        """
        utils.start_continuous_sw_trigger(
            self._session.daq_server,
            self.serial,
            num_triggers=num_triggers,
            wait_time=wait_time,
        )

    @lazy_property
    def max_qubits_per_channel(self) -> int:
        """Maximum number of supported qubits per channel."""
        return utils.max_qubits_per_channel(self._session.daq_server, self.serial)

    @lazy_property
    def qachannels(self) -> t.Sequence[QAChannel]:
        """A Sequence of QAChannels."""
        return NodeList(
            [
                QAChannel(self, self._session, self._tree + ("qachannels", str(i)))
                for i in range(len(self["qachannels"]))
            ],
            self._root,
            self._tree + ("qachannels",),
        )

    @lazy_property
    def scopes(self) -> t.Sequence[SHFScope]:
        """A Sequence of SHFScopes."""
        return NodeList(
            [
                SHFScope(
                    self._root,
                    self._tree + ("scopes", str(i)),
                    self._session.daq_server,
                    self.serial,
                )
                for i in range(len(self["scopes"]))
            ],
            self._root,
            self._tree + ("scopes",),
        )



"""Custom sequence code class."""
from zhinst.toolkit.waveform import Waveforms
import re
import typing as t


class Sequence:
    r"""A representation of a ZI sequencer code.

    This class enables a compact representation of a sequence for a Zurich
    Instruments device. Although a sequencer code can be represented by a
    simple string this class offers the following advantages:

        * Define a constants dictionary. The constants will be added
          automatically to the top of the resulting sequencer code and helps
          to prevent the use of fstrings (which require the escaping of {})
        * Link Waveforms to the sequence. This adds the waveform placeholder
          definitions to the top of the resulting sequencer code.

    Note:
        This class is only for convenience. The same functionality can be
        achieved with a simple string.

    Args:
        code: Sequencer code (default = None).
        constants: A dictionary of constants to be added to the top of the
            resulting sequencer code. (default = None).
        waveforms: Waveforms that will be used in the sequence.

    Example:
        >>> waveforms = Waveforms()
        >>> waveforms[0] = (0.5*np.ones(1008), -0.2*np.ones(1008), np.ones(1008))
        >>> sequencer = Sequence()
        >>> sequencer.constants["PULSE_WIDTH"] = 10e-9 #ns
        >>> sequencer.waveforms = waveforms
        >>> sequencer.code = \"\"\"\
            // Hello World
            repeat(5)
            ...
            \"\"\"
        >>> str(sequencer)
            // Constants
            const PULSE_WIDTH = 10e-9;
            // Waveforms declaration
            assignWaveIndex(placeholder(1008, true, false), placeholder(1008, \
                false, false), 0);
            assignWaveIndex(placeholder(1008, false, false), placeholder(1008, \
                false, false), 2);
            // Hello World
            repeat(5)
            ...
    """

    def __init__(
        self,
        code: str = None,
        *,
        constants: t.Dict[str, float] = None,
        waveforms: Waveforms = None,
    ):
        self._partial_seq = code if code else ""
        self._constants = constants if constants else {}
        self._waveforms = waveforms

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self, *, waveform_snippet: bool = True) -> str:
        """Convert the object into a string.

        Args:
            waveform_snippet: Flag if the waveform declaration should be added
                to the top of the resulting sequence. (default = True).

        Returns:
            String representation of the sequence.
        """
        sequence = self._partial_seq
        if waveform_snippet and self._waveforms:
            sequence = (
                "// Waveforms declaration\n"
                + self._waveforms.get_sequence_snippet()
                + "\n"
                + sequence
            )
        new_constants = {}
        for key, value in self._constants.items():
            constant_regex = re.compile(rf"(const {key} *= *)(.*);")
            if constant_regex.search(sequence):
                sequence = constant_regex.sub(rf"\g<1>{value};", sequence)
            else:
                new_constants[key] = value
        if len(new_constants) > 0:
            sequence = (
                "// Constants\n"
                + "\n".join(
                    [f"const {key} = {value};" for key, value in new_constants.items()]
                )
                + "\n"
                + sequence
            )
        return sequence

    @property
    def code(self):
        """Code of the Sequence."""
        return self._partial_seq

    @code.setter
    def code(self, value):
        """Code of the Sequence."""
        self._partial_seq = value

    @property
    def constants(self):
        """Constants of the Sequence."""
        return self._constants

    @constants.setter
    def constants(self, value):
        """Constants of the Sequence."""
        self._constants = value

    @property
    def waveforms(self):
        """Waveforms of the Sequence."""
        return self._waveforms

    @waveforms.setter
    def waveforms(self, value):
        """Waveforms of the Sequence."""
        self._waveforms = value



"""zhinst-toolkit scope node adaptions for the SHFQA."""
import logging
import typing as t

import zhinst.utils.shfqa as utils
from zhinst.core import ziDAQServer

from zhinst.toolkit.nodetree import Node, NodeTree

logger = logging.getLogger(__name__)


class SHFScope(Node):
    """SHFQA Scope Node.

    Implements basic functionality of the scope node, e.g allowing the user to
    read the data.

    Args:
        root: Root of the nodetree
        tree: Tree (node path as tuple) of the current node
        daq_server: Instance of the ziDAQServer
        serial: Serial of the device.
    """

    def __init__(
        self,
        root: NodeTree,
        tree: tuple,
        daq_server: ziDAQServer,
        serial: str,
    ):
        super().__init__(root, tree)
        self._daq_server = daq_server
        self._serial = serial

    def run(
        self, *, single: bool = True, timeout: float = 10, sleep_time: float = 0.005
    ) -> None:
        """Run the scope recording.

        Args:
            timeout: The maximum waiting time in seconds for the Scope
                (default = 10).
            sleep_time: Time in seconds to wait between requesting the progress
                and records values (default = 0.005).

        Raises:
            TimeoutError: The scope did not start within the specified
                timeout.
        """
        self.single(single)
        self.enable(True)
        try:
            self.enable.wait_for_state_change(1, timeout=timeout, sleep_time=sleep_time)
        except TimeoutError as error:
            raise TimeoutError(
                "Scope could not been started within "
                f"the specified timeout ({timeout})s"
            ) from error

    def stop(self, *, timeout: float = 10, sleep_time: float = 0.005) -> None:
        """Stop the scope recording.

        Args:
            timeout: The maximum waiting time in seconds for the scope
                (default = 10).
            sleep_time: Time in seconds to wait between requesting the progress
                and records values (default = 0.005).

        Raises:
            TimeoutError: The scope did not stop within the specified
                timeout.
        """
        self.enable(False)
        try:
            self.enable.wait_for_state_change(0, timeout=timeout, sleep_time=sleep_time)
        except TimeoutError as error:
            raise TimeoutError(
                "Scope could not been stopped within "
                f"the specified timeout ({timeout})s"
            ) from error

    def wait_done(self, *, timeout: float = 10, sleep_time: float = 0.005) -> None:
        """Wait until the scope recording is finished.

        Args:
            timeout: The maximum waiting time in seconds for the Scope
                (default = 10).
            sleep_time: Time in seconds to wait between requesting the progress
                and records values (default = 0.005).

        Raises:
            TimeoutError: The scope did not finish within the specified
                timeout.
        """
        try:
            self.enable.wait_for_state_change(0, timeout=timeout, sleep_time=sleep_time)
        except TimeoutError as error:
            raise TimeoutError(
                "Scope recording did not finish "
                f"within the specified timeout({timeout})s."
            ) from error

    def configure(
        self,
        *,
        input_select: t.Dict[int, str],
        num_samples: int,
        trigger_input: str,
        num_segments: int = 1,
        num_averages: int = 1,
        trigger_delay: float = 0,
    ) -> None:
        """Configures the scope for a measurement.

        Args:
            input_select: Map of a specific scope channel an their signal
                source, e.g. "channel0_signal_input". (For a list of available
                values use `available_inputs`)
            num_samples: Number samples to recorded in a scope shot.
            trigger_input: Specifies the trigger source of the scope
                acquisition - if set to None, the self-triggering mode of the
                scope becomes active, which is useful e.g. for the GUI.
                For a list of available trigger values use
                `available_trigger_inputs`.
            num_segments: Number of distinct scope shots to be returned after
                ending the acquisition.
            num_averages: Specifies how many times each segment should be
                averaged on hardware; to finish a scope acquisition, the number
                of issued triggers must be equal to num_segments * num_averages.
            trigger_delay: delay in samples specifying the time between the
                start of data acquisition and reception of a trigger.
        """
        utils.configure_scope(
            self._daq_server,
            self._serial,
            input_select=input_select,
            num_samples=num_samples,
            trigger_input=trigger_input,
            num_segments=num_segments,
            num_averages=num_averages,
            trigger_delay=trigger_delay,
        )

    def read(
        self,
        *,
        timeout: float = 10,
    ) -> tuple:
        """Read out the recorded data from the scope.

        Args:
            timeout: The maximum waiting time in seconds for the
                Scope (default: 10).

        Returns:
            (recorded_data, recorded_data_range, scope_time)

        Raises:
            TimeoutError: if the scope recording is not completed before
                timeout.
        """
        return utils.get_scope_data(self._daq_server, self._serial, timeout=timeout)

    @property
    def available_trigger_inputs(self) -> t.List[str]:
        """List of the available trigger sources for the scope."""
        return [
            option.enum for option in self.trigger.channel.node_info.options.values()
        ]

    @property
    def available_inputs(self) -> t.List[str]:
        """List of the available signal sources for the scope channels."""
        return [
            option.enum
            for option in self.channels[0].inputselect.node_info.options.values()
        ]
